import collections.abc
import enum
import random
import lark

try:
    import stix2.parsing as mappings
except ImportError:
    import stix2.core as mappings


def is_tree(node, rule_name=None):
    """
    Determine whether the given parse tree node is a Tree node, and optionally
    verify its type.

    :param node: The node to check
    :param rule_name: A node type, or None if it doesn't matter
    :return: True if the node is a Tree node of the proper type, False
        otherwise.
    """
    return isinstance(node, lark.Tree) and (
        rule_name is None or node.data == rule_name
    )


def is_token(node, token_type=None):
    """
    Determine whether the given parse tree node is a Token node, and optionally
    verify its type.

    :param node: The node to check
    :param token_type: A token type, or None if it doesn't matter
    :return: True if the node is a Token node of the proper type, False
        otherwise.
    """
    return isinstance(node, lark.Token) and (
        token_type is None or node.type == token_type
    )


def rand_iterable(it, len_it=None):
    """
    Choose a uniformly random value from the given iterable.  If len(it) is
    available, len_it may be None.  Otherwise, len_it must be provided as the
    "length" of the given iterable (the number of values which will be
    produced).

    random.choice() seems to require a sequence (needs indexed access), so it
    doesn't work with things like sets.

    :param it: The iterable
    :param len_it: The length of it, or None to obtain the length via len().
    :return: A random value from it
    """
    if len_it is None:
        len_it = len(it)

    for i, val in enumerate(it):
        if random.random() < 1.0 / (len_it - i):
            return val

    raise Exception("Iterable was empty!")


def recurse_references(obj):
    """
    Helper for find_references().  See that function for more information.
    Can also be used as a more generic reference finder, which requires no
    particular structure (e.g. a "type" property) and has no special casing for
    observed-data/objects.

    :param obj: An object.  Can be any type, but values will only be produced
        from mappings.
    """
    if isinstance(obj, collections.abc.Mapping):
        for prop, value in obj.items():
            if prop.endswith("_ref"):
                yield prop, value

            elif prop.endswith("_refs"):
                for ref in value:
                    yield prop, ref

            else:
                yield from recurse_references(value)


def find_references(obj):
    """
    Find all ref prop names and their ID values from the given STIX object.
    This generator generates (ref_prop_name, ID) pairs for all reference
    properties.  For _refs properties (which are list-valued), a separate pair
    is generated for each element of the list; these pairs will contain the
    same property name and a different list element.

    :param obj: A STIX object with a "type" property.
    """
    for prop, value in obj.items():
        if prop.endswith("_ref"):
            yield prop, value

        elif prop.endswith("_refs"):
            for ref in value:
                yield prop, ref

        else:
            # Hack for observed-data: skip the inner SCO graph.  I don't
            # think we ever want to mix the two graphs!
            if obj["type"] != "observed-data" or prop != "objects":
                yield from recurse_references(value)


def recurse_references_assignable(obj):
    """
    Helper for find_references_assignable().  See that function for more
    information.  Can also be used as a more generic reference finder, which
    requires no particular structure (e.g. a "type" property) and has no
    special casing for observed-data/objects.

    :param obj: An object.  Can be any type, but values will only be produced
        from mappings.
    """
    if isinstance(obj, collections.abc.Mapping):
        for prop, value in obj.items():
            if prop.endswith("_ref"):
                yield obj, prop, value, prop

            elif prop.endswith("_refs"):
                for idx, ref_id in enumerate(value):
                    yield value, idx, ref_id, prop

            else:
                yield from recurse_references_assignable(value)


def find_references_assignable(obj):
    """
    Find all ref prop names and their ID values from the given object in a way
    that allows modification of the reference (i.e. assignment of a new ID to
    the reference property).  Generates 4-tuples

    (parent, key, ID, reference property name)

    where parent and key are such that the reference can be modified by
    assignment to the expression "parent[key]".  If the reference property is
    list-valued, parent will be the list and key will be an integer index;
    otherwise, parent will be a mapping and key will be a key from the mapping.
    ID is the value of parent[key] and is provided as a convenience so that
    callers can avoid doing an extra lookup to get it.  The reference property
    name will be the same as key, if key is string-valued (a _ref property
    name).  If parent is a list and key an integer, then the reference property
    name is the key which maps to that list (a _refs property name).

    For list-valued _refs properties, a tuple is generated for each element of
    the list: key and ID will change, but parent and the reference property
    name will be the same.

    :param obj: A STIX object with a "type" property.
    """
    for prop, value in obj.items():
        if prop.endswith("_ref"):
            yield obj, prop, value, prop

        elif prop.endswith("_refs"):
            for idx, ref_id in enumerate(value):
                yield value, idx, ref_id, prop

        else:
            # Hack for observed-data: skip the inner SCO graph.  I don't
            # think we ever want to mix the two graphs!
            if obj["type"] != "observed-data" or prop != "objects":
                yield from recurse_references_assignable(value)


def _stix_type_of(obj_or_type):
    """
    Get a STIX type from the given value: if a string is passed, it is assumed
    to be a STIX type and is returned; otherwise it is assumed to be a mapping
    with a "type" property, and the value of that property is returned.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :return: A STIX type
    """
    if isinstance(obj_or_type, str):
        type_ = obj_or_type
    else:
        type_ = obj_or_type["type"]

    return type_


def _get_stix2_class_maps(stix_version):
    """
    Get the stix2 class mappings for the given STIX version.
    :param stix_version:  A STIX version as a string
    :return: The class mappings.  This will be a dict mapping from some general
        category name, e.g. "object" to another mapping from STIX type
        to a stix2 class.
    """
    stix_vid = "v" + stix_version.replace(".", "")
    cls_maps = mappings.STIX2_OBJ_MAPS[stix_vid]

    return cls_maps


def is_sdo(obj_or_type, stix_version="2.1"):
    """
    Determine whether the given object or type is an SDO.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param stix_version:  A STIX version as a string
    :return: True if the object is an SDO or the type is an SDO type; False
        if not
    """

    # Eventually this needs to be moved into the stix2 library (and maybe
    # improved?); see cti-python-stix2 github issue #450.

    cls_maps = _get_stix2_class_maps(stix_version)
    type_ = _stix_type_of(obj_or_type)
    result = type_ in cls_maps["objects"] and type_ not in {
        "relationship", "sighting", "marking-definition", "bundle",
        "language-content"
    }

    return result


def is_sco(obj_or_type, stix_version="2.1"):
    """
    Determine whether the given object or type is an SCO.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param stix_version:  A STIX version as a string
    :return: True if the object is an SCO or the type is an SCO type; False
        if not
    """
    cls_maps = _get_stix2_class_maps(stix_version)
    type_ = _stix_type_of(obj_or_type)
    result = type_ in cls_maps["observables"]

    return result


def is_sro(obj_or_type, stix_version="2.1"):
    """
    Determine whether the given object or type is an SRO.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param stix_version:  A STIX version as a string
    :return: True if the object is an SRO or the type is an SRO type; False
        if not
    """

    # No STIX version dependence here yet...
    type_ = _stix_type_of(obj_or_type)
    result = type_ in ("sighting", "relationship")

    return result


def is_object(obj_or_type, stix_version="2.1"):
    """
    Determine whether a type is the type of any STIX object.  This includes all
    SDOs, SCOs, meta-objects, and bundle.  If obj_or_type is an object, the
    object's "type" property is checked.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param stix_version:  A STIX version as a string
    :return: True if the (object) type is recognized as a STIX type with
        respect to the given STIX version; False if not
    """
    cls_maps = _get_stix2_class_maps(stix_version)
    type_ = _stix_type_of(obj_or_type)
    result = type_ in cls_maps["observables"] or type_ in cls_maps["objects"]

    return result


class STIXTypeClass(enum.Enum):
    """
    Represents different classes of STIX type.
    """
    SDO = 0
    SCO = 1
    SRO = 2


def is_stix_type(obj_or_type, *types, stix_version="2.1"):
    """
    Determine whether the given object or type satisfies the given constraints.
    'types' must contain STIX types as strings, and/or the STIXTypeClass enum
    values.  STIX types imply an exact match constraint; STIXTypeClass enum
    values imply a more general constraint, that the object or type be in that
    class of STIX type.  These constraints are implicitly OR'd together.

    :param obj_or_type: A mapping with a "type" property, or a STIX type
        as a string
    :param types: A sequence of STIX type strings or STIXTypeClass enum values
    :param stix_version:  A STIX version as a string
    :return: True if the object or type satisfies the constraints; False if not
    """

    for type_ in types:
        if type_ is STIXTypeClass.SDO:
            result = is_sdo(obj_or_type, stix_version)
        elif type_ is STIXTypeClass.SCO:
            result = is_sco(obj_or_type, stix_version)
        elif type_ is STIXTypeClass.SRO:
            result = is_sro(obj_or_type, stix_version)
        else:
            # Assume a string STIX type is given instead of a class enum,
            # and just check for exact match.
            obj_type = _stix_type_of(obj_or_type)
            result = is_object(obj_type, stix_version) and obj_type == type_

        if result:
            break

    else:
        result = False

    return result


def random_generatable_stix_type(
    object_generator, *required_types, stix_version="2.1"
):
    """
    Choose a STIX type at random which satisfies the given type constraints,
    and which the given object generator is able to generate.  See
    is_stix_type() for more discussion on the type constraints.

    :param object_generator: An object generator
    :param required_types: Type constraints, as a sequence of STIX type strings
        and/or STIXTypeClass enum values.  If no types are given, it means no
        types are legal, so None will always be returned.
    :param stix_version: A STIX version as a string
    :return: A STIX type if one could be found which satisfies the given
        constraints; None if one could not be found
    """

    candidate_types = [
        type_ for type_ in object_generator.spec_names
        if is_stix_type(type_, *required_types, stix_version=stix_version)
    ]

    if candidate_types:
        stix_type = random.choice(candidate_types)
    else:
        stix_type = None

    return stix_type


def make_bundle(stix_objs, stix_version):
    """
    Creating a Bundle object of the given spec version, which contains the
    given STIX objects.

    :param stix_objs: The objects the bundle is to contain.  Must be a value
        accepted by the Bundle constructor, e.g. a single or list of objects.
    :param stix_version: A STIX version as a string
    :return: The Bundle object
    """

    cls_maps = _get_stix2_class_maps(stix_version)
    bundle_class = cls_maps["objects"]["bundle"]
    bundle = bundle_class(stix_objs, allow_custom=True)

    return bundle



import stix2
import stix2.base
import stix2.custom
import stix2.properties
import stix2generator
import stix2generator.exceptions
import stix2generator.generation.object_generator


_JSON_SIMPLE_TYPE_STIX_PROPERTY_MAP = {
    "string": stix2.properties.StringProperty,
    "number": stix2.properties.FloatProperty,
    "integer": stix2.properties.IntegerProperty,
    "boolean": stix2.properties.BooleanProperty
}


def _is_list_content_homogenous_type(list_):
    assert list_  # assume the list is not empty

    first_type = stix2generator.generation.object_generator._get_spec_type(
        list_[0]
    )

    result = all(
        stix2generator.generation.object_generator._get_spec_type(elt) == first_type
        for elt in list_[1:]
    )

    return result


def _prop_for_spec(prop_spec, stix_version):
    """
    Given an object generator spec, which is treated as a spec for a property
    value of an object, determine the type of value the spec generates and
    create a stix2 property instance which will accept values of that type.

    :param prop_spec: The object generator spec for a property value
    :param stix_version: The STIX version to use: "2.0", "2.1", etc.
    :return: A stix2 property object
    """
    prop_spec_type = stix2generator.generation.object_generator._get_spec_type(
        prop_spec
    )

    if prop_spec_type in _JSON_SIMPLE_TYPE_STIX_PROPERTY_MAP:
        prop_class = _JSON_SIMPLE_TYPE_STIX_PROPERTY_MAP[prop_spec_type]
        prop_obj = prop_class()

    elif prop_spec_type == "object":
        # DictionaryProperty needs a spec_version parameter.
        prop_obj = stix2.properties.DictionaryProperty(
            spec_version=stix_version
        )

    elif prop_spec_type == "array":
        # For arrays, we need to know the element type.
        if isinstance(prop_spec, list):
            # const array spec.  The list must be non-empty and of homogenous
            # type for us to be able to infer a stix2 property.
            if not prop_spec:
                raise stix2generator.exceptions.EmptyListError()
            if not _is_list_content_homogenous_type(prop_spec):
                raise stix2generator.exceptions.HeterogenousListError(prop_spec)

            element_spec = prop_spec[0]

        else:
            element_spec = prop_spec["items"]

        element_prop_obj = _prop_for_spec(element_spec, stix_version)
        prop_obj = stix2.properties.ListProperty(element_prop_obj)

    else:
        # Maybe we just hit this for "null" specs?
        raise stix2generator.exceptions.IllegalSTIXObjectPropertyType(
            prop_spec_type
        )

    return prop_obj


def stix2_register_custom(spec, obj_type_name, stix_version):
    """
    Dynamically derive a simple Python class for the given object generator
    spec, and register it as a custom SDO with the stix2 library under the
    given name.

    :param spec: An object generator spec for a STIX object, as a dict
    :param obj_type_name: The desired STIX type name to register it as
    :param stix_version: The STIX version to use: "2.0", "2.1", etc.
    """
    spec_type = stix2generator.generation.object_generator._get_spec_type(spec)

    if spec_type != "object":
        raise stix2generator.exceptions.IllegalSTIXObjectSpecType(spec_type)

    if "const" in spec:
        # A const object spec must be a dict with "const" key whose value is
        # a dict.  We will treat every property value in that dict as a const
        # spec.
        const_object = spec["const"]
        prop_specs = {
            prop_name: {"const": prop_value}
            if isinstance(prop_value, dict) else prop_value
            for prop_name, prop_value in const_object.items()
        }
    else:
        prop_specs = spec.get("properties", {})

    prop_map = [
        (prop_name, _prop_for_spec(prop_spec, stix_version))
        for prop_name, prop_spec in prop_specs.items()
    ]

    custom_class = type(
        "AutoRegisteredType_" + obj_type_name,
        (object,), {}
    )

    if stix_version == "2.1":
        stix_decorator = stix2.v21.CustomObject
    else:
        stix_decorator = stix2.v20.CustomObject

    stix_decorator(obj_type_name, prop_map)(custom_class)


def stix2_auto_register_all_custom(specs, stix_version):
    """
    Detect which of the given object generator specs are for "custom" objects,
    try to dynamically derive simple Python classes, and register them with the
    stix2 library as custom SDOs so that using those custom types with the
    prototyping language is possible.

    Custom STIX object specs are detected as:
      - Spec name starts with a capital letter
      - Spec name is not in the builtin object generator spec registry for the
        given STIX version

    :param specs: A mapping from name -> object generator spec
    :param stix_version: The STIX version to use: "2.0", "2.1", etc.
    """
    builtin_registry = stix2generator._get_registry(stix_version)

    custom_object_spec_names = (
        spec_name for spec_name in specs
        if spec_name and spec_name[0].isupper() and
        spec_name not in builtin_registry
    )

    for custom_object_spec_name in custom_object_spec_names:
        custom_object_spec = specs[custom_object_spec_name]
        custom_object_props = custom_object_spec.get('properties', {})
        try:
            stix2_register_custom(
                custom_object_spec,
                custom_object_props.get(
                    'type', custom_object_spec_name.lower()
                ),
                stix_version
            )
        except stix2generator.exceptions.AutoRegistrationError as e:
            if not e.spec_name:
                e.spec_name = custom_object_spec_name

            raise



import collections.abc
import datetime
import logging
import random
import stix2
import stix2.parsing
import stix2.utils
import stix2generator
from stix2generator.exceptions import (
    UnhandledPropertyValueType, SpecificationNotFoundError,
    UnsupportedObjectStructureError, UnrecognizedSTIXTypeError,
    InvalidRefPropertyValueError
)
import stix2generator.generation
import stix2generator.utils


# Don't create object paths which include the following properties.
_OBJECT_PATH_BLACKLIST = {
    "id",
    "type",
    "spec_version"
}


# Maps types we would find as property values of stix2 objects, to the
# operators we may use them with in a comparison expression.
_PYTHON_TYPE_TO_OPERATORS = {
    int: ("<", "<=", ">", ">=", "=", "!="),
    float: ("<", "<=", ">", ">=", "=", "!="),
    str: ("<", "<=", ">", ">=", "=", "!="),
    stix2.utils.STIXdatetime: ("<", "<=", ">", ">=", "=", "!="),
    bool: ("=", "!="),
}


# stix2's AST uses a different class for each comparison operator.  Use this
# to resolve an operator to the corresponding AST class.
_OPERATOR_TO_AST_CLASS = {
    "<": stix2.LessThanComparisonExpression,
    "<=": stix2.LessThanEqualComparisonExpression,
    ">": stix2.GreaterThanComparisonExpression,
    ">=": stix2.GreaterThanEqualComparisonExpression,
    "=": stix2.EqualityComparisonExpression,
    "!=": stix2.EqualityComparisonExpression,
    # "!=" uses the same EqualityComparisonExpression class, just constructed
    # differently, so it will have to be a special case.
}


# Represents list index "star" steps in object paths
_INDEX_STAR_STEP = object()


class Config(stix2generator.generation.Config):
    """
    STIX pattern generator config settings.

    min/max_pattern_size: Bounds on pattern size, in terms of the total number
        of simple comparison expressions across all observation expressions.

    min/max_repeat_count: bounds on the repeat count for the REPEATS qualifier.

    min/max_within_count: bounds on the number of seconds for the WITHIN
        qualifier.

    probability_qualifier: Probability a given observation expression (at any
        nesting level) will get a random qualifier.

    probability_continue_path_through_ref: When randomly generating an object
        path, determines how likely the path will continue through a reference
        property to another object.

    probability_index_star_step: When randomly generating an object path,
        determines how likely list index steps will use '*'.
    """
    _DEFAULTS = {
        "min_pattern_size": 1,
        "max_pattern_size": 5,
        "min_repeat_count": 1,
        "max_repeat_count": 10,
        "min_within_count": 1,
        "max_within_count": 10,
        "probability_qualifier": 0.2,
        "probability_continue_path_through_ref": 0.8,
        "probability_index_star_step": 0.1
    }


def _rand_series(n):
    """
    Generate a random series (sum) which adds to 'n'.  The generated series
    takes the form of a sequence of positive integers which add to n.  If
    n <= 0, nothing is generated.

    :param n: The desired sum, as an integer
    """
    # With this algorithm, the probability distribution over series is not
    # uniform: it favors shorter series over longer ones.  Should I try to
    # think of something better?
    while n > 0:
        k = random.randint(1, n)
        yield k
        n -= k


def _random_operator_for_type(type_):
    """
    Pick a random comparison expression "operator" for the given type.  In
    fact, there is a different AST class for (most) comparison expression
    operators, so we must actually choose a random class instead.

    There is no class corresponding to "!="; it is realized with the class for
    equality, and with a flag for negation.  So we must actually return both a
    class and a boolean indicating whether to negate.

    :param type_: A type, e.g. int, str, etc
    :return: A 2-tuple consisting of (1) an AST class, and (2) a boolean
        negation value, or None if the type is unrecognized and a set of legal
        operators can't be determined.
    """
    candidate_ops = _PYTHON_TYPE_TO_OPERATORS.get(type_)
    if not candidate_ops:
        return None

    op = random.choice(candidate_ops)
    ast_class = _OPERATOR_TO_AST_CLASS[op]

    # Special case... there is no AST class for not-equal.  Instead, it uses
    # EqualityComparisonExpression with a flag for negation.
    negated = op == "!="

    return ast_class, negated


def _is_ref_path(path_elements):
    """
    Determine whether the given object path, expressed as an element list
    (see _element_list_to_object_path()), ends with a reference and is
    therefore eligible for continuation through the reference.  The given
    object path is assumed to be "completed" down to a single STIX property
    value.  This means that a *_ref property will be the last component, and
    *_refs will be second-to-last, because it requires a subsequent index step.

    :param path_elements: An object path, as a list
    :return: True if a continuable reference path; False if not
    """
    result = False
    if path_elements:
        last_elt = path_elements[-1]
        if isinstance(last_elt, str) and last_elt.endswith("_ref"):
            result = True

        elif len(path_elements) > 1:
            # for _refs properties, the ref property itself must be
            # second-to-last, and the last path element must be an index step,
            # either "*" or an int.  Maybe not necessary to check the index
            # step; all we need is to check the second-to-last property.
            second_last_elt = path_elements[-2]
            if isinstance(second_last_elt, str) \
                    and second_last_elt.endswith("_refs"):
                result = True

    return result


def _element_list_to_object_path(object_type, path_elements):
    """
    Build an AST ObjectPath instance from an object "path" given as a list
    of strings, ints and the special _INDEX_STAR_STEP object, used for list
    index "star" steps.  The strings are interpreted as property names and
    the ints/star steps as list indices.

    :param object_type: The SCO type to use for the ObjectPath instance
    :param path_elements: The path elements as a list of
        strings/ints/_INDEX_STAR_STEPs
    :return: An ObjectPath instance
    """

    path_components = []
    i = 0
    while i < len(path_elements):
        elt_i = path_elements[i]

        if not isinstance(elt_i, str):
            raise UnsupportedObjectStructureError(
                object_type, path_elements
            )

        if i < len(path_elements) - 1:
            elt_i1 = path_elements[i+1]
            if isinstance(elt_i1, int):
                component = stix2.ListObjectPathComponent(elt_i, elt_i1)
                i += 1

            elif elt_i1 is _INDEX_STAR_STEP:
                component = stix2.ListObjectPathComponent(elt_i, "*")
                i += 1

            # ignoring ReferenceObjectPathComponent here.  I think the pattern
            # visitor never uses it(?), so I guess I won't either.

            else:
                component = stix2.BasicObjectPathComponent(elt_i, False)

        else:
            component = stix2.BasicObjectPathComponent(elt_i, False)

        path_components.append(component)
        i += 1

    object_path = stix2.ObjectPath(object_type, path_components)

    return object_path


class PatternGenerator:
    """
    Instances of this class generate random STIX patterns.
    """

    def __init__(self, object_generator, stix_version, config=None):
        """
        Initialize this PatternGenerator.  Patterns are generated by randomly
        walking through randomly generated objects, so a pattern generator
        relies on an object generator for its random STIX content.

        :param object_generator: A STIX object generator
        :param stix_version: The STIX version to generate patterns for.  (This
            should probably match up with the version of objects generated by
            object_generator!)
        :param config: A Config object with settings for pattern generation,
            or None to choose default settings
        """
        cls = self.__class__
        self.__log = logging.getLogger(
            cls.__module__ + "." + cls.__name__
        )

        self.__generator = object_generator
        self.__stix_version = stix_version
        self.__config = config or Config()

    def __random_sco_type(self):
        return stix2generator.utils.random_generatable_stix_type(
            self.__generator,
            stix2generator.utils.STIXTypeClass.SCO,
            stix_version=self.__stix_version
        )

    def __generate_object_path(self, type_constraint=None):
        """
        Generate a random object path.  This is done by generating a random
        object, and then choosing a random path through it.  If type_constraint
        is given, an object of that type is generated.  Otherwise, a random SCO
        type is chosen.  The value at the "end" of the path is also returned,
        for use in the pattern.

        :param type_constraint: An SCO type, or None
        :return: A 2-tuple consisting of (1) the ObjectPath instance, and (2)
            a value from the object.  This value will be taken from a stix2
            object, so that determines its type.  It could be a string,
            STIXdatetime instance, etc.
        """
        if type_constraint:
            sco_type = type_constraint
        else:
            sco_type = self.__random_sco_type()

        try:
            obj_dict = self.__generator.generate(sco_type)
        except SpecificationNotFoundError as e:
            raise UnrecognizedSTIXTypeError(sco_type) from e

        obj = stix2.parse(obj_dict, allow_custom=True)

        path_elements = []
        while True:
            if isinstance(obj, collections.abc.Mapping):
                candidate_props = obj.keys() - _OBJECT_PATH_BLACKLIST
                element = stix2generator.utils.rand_iterable(candidate_props)
            elif isinstance(obj, list):
                element = random.randrange(len(obj))
            else:
                break

            # Let's have a chance to append an index "star" step when a
            # list is encountered, instead of the chosen index.
            if isinstance(element, int) and \
                    random.random() < self.__config.probability_index_star_step:
                path_elements.append(_INDEX_STAR_STEP)
            else:
                path_elements.append(element)

            obj = obj[element]

        object_path = _element_list_to_object_path(sco_type, path_elements)

        if _is_ref_path(path_elements) and \
                random.random() < \
                self.__config.probability_continue_path_through_ref:

            # If a ref path, the value must be an ID.  Extract the object type
            # from the ID and generate a path of that type to concatenate to our
            # path.  In this way, we can continue the path through references.
            dd_idx = obj.find("--")
            if dd_idx == -1:
                raise InvalidRefPropertyValueError(obj)

            id_type = obj[:dd_idx]

            try:
                path_continuation, obj = self.__generate_object_path(id_type)
            except UnrecognizedSTIXTypeError:
                # We couldn't generate an SCO of type id_type.  Reduce this to
                # a warning; we will simply not have a continued path in this
                # case.
                self.__log.warning(
                    'Truncating object path due to unrecognized SCO type "%s"',
                    id_type
                )

            else:
                object_path.property_path.extend(
                    path_continuation.property_path
                )

        return object_path, obj

    def __generate_simple_comparison_expression(self, type_constraint=None):
        """
        Generate a "simple" <path> <op> <value> comparison expression.  If
        a type constraint is given, that will be the SCO type for the path.
        Otherwise, a random SCO type is chosen.

        :param type_constraint: An SCO type, or None
        :return: An AST instance for a simple comparison expression
        """

        object_path, value = self.__generate_object_path(type_constraint)

        result = _random_operator_for_type(type(value))
        if result is None:
            raise UnhandledPropertyValueType(value)

        ast_class, negated = result
        ast_node = ast_class(object_path, value)
        ast_node.negated = negated

        return ast_node

    def __generate_simple_comparison_expression_list(
        self, size, type_constraint, is_and
    ):
        """
        Generate a list of the given size of "simple" comparison expressions,
        which honors the given type constraint, relative to the indicated
        boolean connective.  is_and indicates how the returned comparison
        expressions will be used.  They will be connected with 'AND' if is_and
        is True, else 'OR'.  Therefore, if a type constraint is given and
        is_and is True, all generated comparison expressions must be of the
        given type.  Otherwise, at least one must be of the given type.  If no
        type constraint is given then is_and must be False, because AND'd
        comparison expressions require a constraint.  If is_and is False and
        no type constraint is given, all comparison expressions will be of
        randomly chosen types.

        :param size: The number of simple comparison expressions to generate
        :param type_constraint: An SCO type, or None
        :param is_and: True if the returned expressions will be connected via
            AND; False if they will be connected via OR.
        :return: The list of comparison expressions (ASTs).
        """
        assert size >= 0

        # If AND, all operands *must* be type-constrained.
        assert not is_and or type_constraint

        if type_constraint:
            if is_and:
                # In 'AND': all simple exprs must be of the same type
                result = [
                    self.__generate_simple_comparison_expression(
                        type_constraint
                    )
                    for _ in range(size)
                ]

            else:
                # In 'OR': at least one must be of the constraining type.

                # Create N-1 unconstrained exprs and 1 constrained expr...
                if size == 0:
                    result = []

                else:
                    result = [
                        self.__generate_simple_comparison_expression(None)
                        for _ in range(size-1)
                    ]

                    constrained_expr = \
                        self.__generate_simple_comparison_expression(
                            type_constraint
                        )

                    # Then insert the constrained one at a random location in
                    # the list
                    result.insert(
                        random.randint(0, len(result)),
                        constrained_expr
                    )

        else:
            # no type constraint; must be an 'OR'.  So we can generate whatever
            # types we want.
            result = [
                self.__generate_simple_comparison_expression(
                    None
                )
                for _ in range(size)
            ]

        return result

    def __generate_complex_comparison_expression(
        self, size, type_constraint=None
    ):
        """
        Generates a "complex" comparison expression, i.e. one which may consist
        of sub-expressions connected via AND or OR.  If a type constraint is
        given, the resulting expression will honor that constraint.

        :param size: The size of the desired complex comparison expression, in
            terms of the number of simple comparison expressions it must contain
        :param type_constraint: An SCO type, or None
        :return:
        """
        assert size > 0

        # This complex expression must be composed of N simple expressions.
        # This implementation builds the overall expression in two parts: a
        # left and right side.  The location of the split between left and
        # right is random.  A side is randomly chosen to just contain a series
        # of simple expressions, and the other side will have a nested
        # subexpression.
        #
        # One goal of the strategy is to avoid excessive nested parentheses.
        # Too many parentheses results in ugly crazy-looking patterns.  This
        # algorithm still can generate some silly patterns, but I hope it helps
        # a little.
        if size == 1:
            expr = self.__generate_simple_comparison_expression_list(
                1, type_constraint, False
            )[0]

        else:

            # Choose whether top-level operator will be AND or OR.
            # This will also determine how we handle the type constraint.
            is_and = random.random() < 0.5

            # If AND, all operands *must* be type-constrained.
            if is_and and not type_constraint:
                type_constraint = self.__random_sco_type()

            # In the following, if type_constraint is None, both left and right
            # constraints will be None.  No need for a special case.  If we
            # have a type constraint, for 'AND', the constraint must be
            # enforced on both sides.  For 'OR', we need only enforce it on one
            # side.
            if is_and:
                left_constraint = right_constraint = type_constraint
            else:
                left_constraint, right_constraint = type_constraint, None
                if random.random() < 0.5:
                    left_constraint, right_constraint = \
                        right_constraint, left_constraint

            # Don't let either side be zero size here.  Avoids the case where
            # we have an OR, and randomly choose to enforce the type constraint
            # on the zero-length side.  That can result in an invalid pattern.
            lsize = random.randint(1, size-1)
            rsize = size - lsize

            if random.random() < 0.5:
                # Parenthesize right case
                operands = self.__generate_simple_comparison_expression_list(
                    lsize, left_constraint, is_and
                )

                operands.append(stix2.ParentheticalExpression(
                    self.__generate_complex_comparison_expression(
                        rsize, right_constraint
                    )
                ))

            else:
                # Parenthesize left case
                operands = [stix2.ParentheticalExpression(
                    self.__generate_complex_comparison_expression(
                        lsize, left_constraint
                    )
                )]

                operands.extend(
                    self.__generate_simple_comparison_expression_list(
                        rsize, right_constraint, is_and
                    )
                )

            if is_and:
                expr = stix2.AndBooleanExpression(operands)
            else:
                expr = stix2.OrBooleanExpression(operands)

        return expr

    def __generate_random_qualifier(self):
        """
        Generate a random qualifier AST object.

        :return: The qualifier object
        """
        qual_type = random.randrange(3)

        if qual_type == 0:
            repeat_count = random.randint(
                self.__config.min_repeat_count,
                self.__config.max_repeat_count
            )
            qualifier = stix2.RepeatQualifier(repeat_count)

        elif qual_type == 1:
            within_count = random.randint(
                self.__config.min_within_count,
                self.__config.max_within_count
            )
            qualifier = stix2.WithinQualifier(within_count)

        else:
            # Let's make the random timestamps near the current time
            # (within a year).
            dur1 = datetime.timedelta(
                microseconds=random.randrange(
                    # 1 year
                    1000000 * 60 * 60 * 24 * 365
                )
            )

            dur2 = datetime.timedelta(
                microseconds=random.randrange(
                    # 1 year
                    1000000 * 60 * 60 * 24 * 365
                )
            )

            if random.random() < 0.5:
                dur1 = -dur1

            if random.random() < 0.5:
                dur2 = -dur2

            now = datetime.datetime.utcnow()
            dt1 = now + dur1
            dt2 = now + dur2

            # Order them as start=dt1, stop=dt2
            if dt1 > dt2:
                dt1, dt2 = dt2, dt1

            elif dt1 == dt2:
                # in the remote chance we get the same timestamp for both,
                # just nudge one ahead...
                dt2 += datetime.timedelta(seconds=1)

            # STIX 2.0 requires string constants and millisecond precision
            # here...
            if self.__stix_version == "2.0":
                dt1_str = stix2.utils.format_datetime(
                    stix2.utils.STIXdatetime(dt1, precision="millisecond")
                )
                dt1 = stix2.patterns.StringConstant(dt1_str)

                dt2_str = stix2.utils.format_datetime(
                    stix2.utils.STIXdatetime(dt2, precision="millisecond")
                )
                dt2 = stix2.patterns.StringConstant(dt2_str)

            qualifier = stix2.StartStopQualifier(dt1, dt2)

        return qualifier

    def __generate_observation_expression(self, size):
        """
        Generate a random complex observation expression, which may consist of
        sub-expressions and qualifiers.

        :param size: The size of the desired observation expression, in terms of
            the number of simple comparison expressions it must contain
        :return: The observation expression AST
        """
        assert size > 0

        # The generation strategy is similar to that for comparison expressions
        # (see __generate_complex_comparison_expression()).  It is generated in
        # two parts of random size; one side is constructed as a sub-expression.

        if size == 1:
            obs_expr = stix2.ObservationExpression(
                self.__generate_complex_comparison_expression(1)
            )

        else:
            lsize = random.randint(0, size)
            rsize = size - lsize

            if random.random() < 0.5:
                # Parenthesize right case
                obs_exprs = [
                    stix2.ObservationExpression(
                        self.__generate_complex_comparison_expression(sz)
                    )
                    for sz in _rand_series(lsize)
                ]

                if rsize > 0:
                    obs_exprs.append(stix2.ParentheticalExpression(
                        self.__generate_observation_expression(rsize)
                    ))

            else:
                # Parenthesize left case
                if lsize == 0:
                    obs_exprs = []
                else:
                    obs_exprs = [stix2.ParentheticalExpression(
                        self.__generate_observation_expression(lsize)
                    )]

                obs_exprs.extend(
                    stix2.ObservationExpression(
                        self.__generate_complex_comparison_expression(sz)
                    )
                    for sz in _rand_series(rsize)
                )

            ast_class = random.choice((
                stix2.AndObservationExpression,
                stix2.OrObservationExpression,
                stix2.FollowedByObservationExpression
            ))

            obs_expr = ast_class(obs_exprs)

        if random.random() < self.__config.probability_qualifier:
            qualifier = self.__generate_random_qualifier()
            obs_expr = stix2.QualifiedObservationExpression(obs_expr, qualifier)

        return obs_expr

    def generate_ast(self):
        """
        Generate a random STIX pattern as an AST.

        :return: A pattern AST
        """

        size = random.randint(
            self.__config.min_pattern_size,
            self.__config.max_pattern_size
        )

        return self.__generate_observation_expression(size)

    def generate(self):
        """
        Generate a random STIX pattern.

        :return: A pattern string
        """

        pattern_ast = self.generate_ast()
        return str(pattern_ast)



import argparse
import math

import stix2generator
import stix2generator.generation.object_generator
import stix2generator.generation.reference_graph_generator
import stix2generator.generation.stix_generator
import stix2generator.logging
import stix2generator.utils


def nonnegative_int(arg):
    """
    Verify/convert arg to a non-negative integer.

    :param arg: A commandline arg as a string
    :return: The integer
    :raises argparse.ArgumentTypeError: If the arg can't be converted or is
        out of range
    """
    try:
        int_arg = int(arg)
    except ValueError as e:
        raise argparse.ArgumentTypeError(
            "Not a nonnegative integer: {}".format(arg)
        ) from e

    if int_arg < 0:
        raise argparse.ArgumentTypeError(
            "Not a nonnegative integer: {}".format(arg)
        )

    return int_arg


def positive_int(arg):
    """
    Verify/convert arg to a positive integer.

    :param arg: A commandline arg as a string
    :return: The integer
    :raises argparse.ArgumentTypeError: If the arg can't be converted or is
        out of range
    """
    try:
        int_arg = int(arg)
    except ValueError as e:
        raise argparse.ArgumentTypeError(
            "Not a positive integer: {}".format(arg)
        ) from e

    if int_arg < 1:
        raise argparse.ArgumentTypeError(
            "Not a positive integer: {}".format(arg)
        )

    return int_arg


def probability(arg):
    """
    Verify/convert arg to a probability value in the range [0, 1].

    :param arg: A commandline arg as a string
    :return: The probability as a float
    :raises argparse.ArgumentTypeError: If the arg can't be converted or is
        out of range
    """
    try:
        float_arg = float(arg)
    except ValueError as e:
        raise argparse.ArgumentTypeError(
            "Not a float in [0,1]: {}".format(arg)
        ) from e

    if math.isnan(float_arg) or not 0 <= float_arg <= 1:
        raise argparse.ArgumentTypeError(
            "Not a float in [0,1]: {}".format(arg)
        )

    return float_arg


def parse_args():
    arg_parser = argparse.ArgumentParser(
        description="Generation random STIX content"
    )

    arg_parser.add_argument(
        "--min-rels",
        help="""Minimum number of SROs to create.  Default=%(default)d""",
        type=positive_int,
        default=1
    )

    arg_parser.add_argument(
        "--max-rels",
        help="""Maximum number or SROs to create.  Default=%(default)d""",
        type=positive_int,
        default=5
    )

    arg_parser.add_argument(
        "--p-reuse",
        help="""
        Probability of object reuse, when creating new connections among
        objects.  Must be a real number in [0, 1].  Lower values result in a
        graph with more nodes and less interconnection.  Higher values result
        in a graph with fewer nodes and more interconnection.
        Default=%(default).1f
        """,
        type=probability,
        default=0.5
    )

    arg_parser.add_argument(
        "--p-sighting",
        help="""
        Probability that when an SRO is added, it is a sighting.  Must be a
        real number in [0, 1].  Default=%(default).1f
        """,
        type=probability,
        default=0.1
    )

    arg_parser.add_argument(
        "--dangling-refs",
        help="""
        Leave reference properties "dangling".  Don't force them to refer to
        existing objects.  Applies to all reference properties *except* the
        endpoints of SROs.
        """,
        action="store_true"
    )

    arg_parser.add_argument(
        "--ref-max-depth",
        help="""
        If creating a new object to avoid a dangling reference, the new object
        could itself have reference properties; new objects created to satisfy
        those could themselves have reference properties, etc.  This setting
        limits how far we grow this "reference graph".  Enforcement of this
        limit is best-effort; reference properties required by the
        specification may cause further growth.  Only applicable if
        --dangling-refs is not given.  Must be a non-negative integer.
        Default=%(default)d
        """,
        type=nonnegative_int,
        default=0
    )

    arg_parser.add_argument(
        "-v", "--verbose",
        help="""
        Enable verbose diagnostic output.  Repeat for
        increased verbosity.
        """,
        action="count"
    )

    arg_parser.add_argument(
        "--stix-version",
        help="STIX version to use.  Default=%(default)s",
        choices=["2.0", "2.1"],
        default="2.1"
    )

    arg_parser.add_argument(
        "-b", "--bundle",
        help="Create a bundle",
        action="store_true"
    )

    args = arg_parser.parse_args()

    return args


def main():
    args = parse_args()

    stix2generator.logging.config_logging(args.verbose)

    obj_gen_config = stix2generator.generation.object_generator.Config(
        # hard-code; we need this disabled for ref graph max-depth setting
        # to mean anything.
        minimize_ref_properties=False
    )

    ref_gen_config = stix2generator.generation.reference_graph_generator.Config(
        max_depth=args.ref_max_depth,
        probability_reuse=args.p_reuse
    )

    stix_gen_config = stix2generator.generation.stix_generator.Config(
        min_relationships=args.min_rels,
        max_relationships=args.max_rels,
        probability_reuse=args.p_reuse,
        probability_sighting=args.p_sighting,
        complete_ref_properties=not args.dangling_refs
    )

    stix_gen = stix2generator.create_stix_generator(
        object_generator_config=obj_gen_config,
        ref_graph_generator_config=ref_gen_config,
        stix_generator_config=stix_gen_config,
        stix_version=args.stix_version
    )

    graph = stix_gen.generate()

    if args.bundle:
        bundle = stix2generator.utils.make_bundle(
            list(graph.values()), args.stix_version
        )

        print(bundle.serialize(pretty=True))

    else:
        for obj in graph.values():
            print(obj.serialize(pretty=True))


if __name__ == "__main__":
    main()



import logging
import logging.config


# custom more-verbose-than-debug level, for extra verbosity.
EXTRA_VERBOSE = 5


def _detect_top_level_module_name():
    """
    Simple function to try to auto-detect the top-level module name, based
    on the name of this module.  This can be used to name a logger.  Hopefully
    this makes it adapt to module refactoring; we don't have to hard-code
    anything.  I anticipate that all logging in this library will be done via
    loggers named after the module hierarchy, so we can affect all of them
    as a group this way.

    :return: The top level module name
    """
    dot_idx = __name__.find(".")

    if dot_idx == -1:
        top_level_module = __name__
    else:
        top_level_module = __name__[:dot_idx]

    return top_level_module


def config_logging(verbosity_count):
    """
    Configure logging for commandline scripts.  Library code should not
    configure logging.

    :param verbosity_count: The count of "-v"s on the commandline, or None
    """

    if verbosity_count is None or verbosity_count == 0:
        log_level = logging.INFO
    elif verbosity_count == 1:
        log_level = logging.DEBUG
    else:
        log_level = EXTRA_VERBOSE

    config = {
        "version": 1,

        "formatters": {
            "message_only": {
                "format": "%(message)s"
            }
        },

        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "message_only"
            }
        },

        "disable_existing_loggers": False
    }

    logger_config = {
        "level": log_level,
        "handlers": ["console"]
    }

    # In addition to changing the log level, if extra verbosity is
    # selected, configure the root handler so that all libs will show
    # logging.  Otherwise, just configure a logger for this library.
    if log_level == EXTRA_VERBOSE:
        config["root"] = logger_config

    else:
        top_level_module = _detect_top_level_module_name()
        config["loggers"] = {
            top_level_module: logger_config
        }

    logging.config.dictConfig(config)



import lark
import logging
import pprint
import stix2

from stix2generator.exceptions import (
    CircularVariableDependenciesError, LanguageError, RedeclaredVariableError,
    UndeclaredVariableError
)
from stix2generator.utils import (
    is_token, is_tree
)
import stix2generator.generation.semantics
import stix2generator.generation.object_generator
import stix2generator.logging


_grammar = r"""

// can we use the Lark-included string literal token?
%import common.ESCAPED_STRING

graph_spec : (statement ".")+

statement : variable_declaration_statement
  | graph_statement
  | sighting

variable_declaration_statement : variable_declaration ("," variable_declaration)* ":" SDO_TYPE_NAME

variable_declaration : count? VARIABLE_NAME property_block?

graph_statement : (sdo_ref | sdo_list) relationship?

relationship : count? RELATIONSHIP_NAME graph_statement

sdo_list : "(" sdo_ref+ ")"

sdo_ref : sdo_inline | VARIABLE_NAME

sdo_inline : count? SDO_TYPE_NAME property_block?

property_block : "{" [property_assignment ("," property_assignment)*] "}"

property_assignment : PROPERTY_NAME ":" (graph_statement | ESCAPED_STRING | string_array)

count : POSITIVE_INT

string_array : "[" [ESCAPED_STRING ("," ESCAPED_STRING)*] "]"

// This allows both the property block and "of" clause to be omitted
// (which will result in an error since "sighting_of_ref" is a required
// property), but it's simpler than expressing "at least one".
// Increase rule priority, since "Sighting of ..." is ambiguous w.r.t.
// graph_statement, with (supposed) sdo "Sighting" and relationship "of".
sighting.2 : "Sighting" property_block? ["of" graph_statement]

POSITIVE_INT : /[1-9][0-9]*/

SDO_TYPE_NAME : /[A-Z][A-Za-z0-9_]*/

// These definitions conflict; hopefully we can rely on Lark's
// terminal collision resolution.
RELATIONSHIP_NAME : /[a-z][a-z0-9-]*/
PROPERTY_NAME : /[a-z][a-z0-9_]*/
VARIABLE_NAME : /[a-z][a-z0-9_-]*/

WS : /[ \t\r\n\u000B\u000C\u0085\u00a0\u1680\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000]+/

%ignore WS

"""


_parser = lark.Lark(
    _grammar,
    start="graph_spec",
    propagate_positions=True,
    # debug=True
)


# lazy-initialized
_log = None


# A custom property to use when embedding variable names in objects
_VARIABLE_NAME_PROPERTY = "x_viz_variable_name"


def _get_logger():
    global _log
    if _log is None:
        _log = logging.getLogger(__name__)

    return _log


def _make_object(object_generator, type_name, overlay_props=None):
    """
    Make an object via an object generator.

    :param object_generator: The generator to use
    :param type_name: The name of a spec, which the generator should recognize
    :param overlay_props: After having generated the object, a dict of
        property names and values which will be overlaid.  This can act to
        override generated property values.
    :return: The object, as a dict
    """

    sdo_dict = object_generator.generate(type_name)

    if overlay_props:
        sdo_dict.update(overlay_props)

    return sdo_dict


def _make_sro(object_generator, source_id, rel_type, target_id):
    """
    Make an SRO.

    :param source_id: The ID of the source object (string)
    :param rel_type: The relationship type (string)
    :param target_id: The ID of the target object (string)
    :return: The SRO, as a dict
    """

    rel = object_generator.generate("relationship")
    rel["source_ref"] = source_id
    rel["target_ref"] = target_id
    rel["relationship_type"] = rel_type

    return rel


def _make_sighting(object_generator, sighting_of, overlay_props=None):
    """
    Make a Sighting.

    :param sighting_of: A value for the sighting_of_ref property, or None to
        omit that property.
    :param overlay_props: After having generated the object, a dict of
        property names and values which will be overlaid.  This can act to
        override generated property values.
    :return: The sighting, as a dict
    """

    rel = object_generator.generate("sighting")

    if sighting_of:
        rel["sighting_of_ref"] = sighting_of
    if overlay_props:
        rel.update(overlay_props)

    return rel


def _print_parse_tree(logger, tree, indent=0):
    """
    A parse tree printer that's better than Lark's (in my opinion).  It shows
    token types as well as values (Lark's doesn't show the types).  It's done
    via the logger, at a custom extra-verbosity level.

    :param logger: The logger to use
    :param tree: The parse tree
    :param indent: An indent amount
    """
    line = " " * indent
    if is_token(tree):
        line += tree + " (" + tree.type + ")"
        logger.log(stix2generator.logging.EXTRA_VERBOSE, line)
    else:
        line += tree.data
        logger.log(stix2generator.logging.EXTRA_VERBOSE, line)
        for child in tree.children:
            _print_parse_tree(logger, child, indent + 1)


def _string_token_value(tok):
    """
    Given an ESCAPED_STRING token, unquote and unescape its value, to
    obtain the actual string it represents.

    :param tok: an ESCAPED_STRING token
    :return: The string value
    """
    return tok.value[1:-1].replace('\\"', '"').replace("\\\\", "\\")


def _topo_sort_dfs(dependencies, start_var, visited_vars, sorted_vars,
                   search_path):
    """
    Helper for topo sort, which does a post-order DFS from the given start
    variable to find the ordering.

    :param dependencies: The dependency data
    :param start_var: The start variable for the search
    :param visited_vars: Keeps track of which variables we've already seen,
        to prevent duplicates
    :param sorted_vars: A list which builds up the sort order
    :param search_path: A stack which records our DFS search path, so that
        we can detect cycles
    """

    visited_vars.add(start_var)
    search_path.append(start_var)

    if start_var in dependencies:
        for dep_var in dependencies[start_var]:

            if dep_var in search_path:
                cyclic_path = search_path[search_path.index(dep_var):]
                cyclic_path.append(dep_var)
                raise CircularVariableDependenciesError(cyclic_path)

            if dep_var not in visited_vars:
                _topo_sort_dfs(
                    dependencies, dep_var, visited_vars, sorted_vars,
                    search_path
                )

                sorted_vars.append(dep_var)

    search_path.pop()


def _topo_sort_dependencies(dependencies):
    """
    Topologically sorts the variables given in dependencies, to obtain a
    value creation order that is compatible with everyone's dependencies.

    :param dependencies: The dependency data, as a mapping from var name to
        a set of dependencies.  (This is basically an adjacency list
        representation of a directed graph.)
    :return: A list of variable names in sorted order
    :raises CircularVariableDependenciesError: If circular dependencies are
        encountered
    """
    visited_vars = set()
    sorted_vars = []

    for var_name in dependencies:
        if var_name not in visited_vars:
            _topo_sort_dfs(
                dependencies, var_name, visited_vars, sorted_vars, []
            )
            sorted_vars.append(var_name)

    return sorted_vars


def _remove_variable_declaration_statements(parse_tree):
    """
    Modify the parse tree by removing all variable declaration statements.
    This is an in-place modification: the given tree's list of children is
    modified.

    :param parse_tree: The whole parse tree (a subtree won't work)
    """

    indices_to_remove = [
        i for i, statement in enumerate(parse_tree.children)
        if statement.children[0].data == "variable_declaration_statement"
    ]

    # Easier to do in reverse order, so that deletions don't affect
    # subsequent indices.
    for i in reversed(indices_to_remove):
        del parse_tree.children[i]


class _VariableSpec:
    """
    Instances of this class serve as placeholders for variables which have
    dependencies.  They contain the information necessary to create a variable
    value later.  Their values aren't created when their declarations are
    encountered in the parse tree.
    """
    def __init__(self, var_name, var_count, var_type, var_props_tree):
        self.var_name = var_name
        self.var_count = var_count
        self.var_type = var_type
        self.var_props_tree = var_props_tree

    def __str__(self):
        s = "Placeholder for '{}', count={}, type={}".format(
            self.var_name, self.var_count, self.var_type
        )

        return s


class _VariableDependencyCollector(lark.Visitor):
    """
    A visitor which is intended to operate *only* on a variable property
    block.  It simply collects all variable names from the block.  These
    are the dependencies of the variable.  Note that there can never be
    indirect dependencies, since variables are never allowed to have property
    blocks, except when they are declared (i.e. no variable inside the block
    can itself have a block).  This makes it really easy: just grab all
    variable names you can find.
    """

    def __init__(self):
        self.variables = set()

    def sdo_ref(self, tree):
        if is_token(tree.children[0], "VARIABLE_NAME"):
            self.variables.add(tree.children[0])


class _VariableProcessor(lark.Transformer):
    """
    Selectively processes just the variable declarations in the parse tree.
    This creates STIX objects for all variables without any dependencies
    (references to other variables in a property block).  For variables with
    dependencies, objects are not created here.  Instead, a placeholder is
    created, and the dependency data is stored in a mapping to keep track of
    it.  Those objects are created later, since they can't be created as they
    are encountered in this parse tree traversal.
    """

    def __init__(self, object_generator):
        """
        Initialize the instance.

        :param object_generator: The object generator which will be used to
            generate values of variables.
        """
        super().__init__()

        self.__object_generator = object_generator

        self.__variables = {}
        self.__objects = {}
        self.__dependencies = {}

    def graph_spec(self, children):
        """
        Callback for the top-level rule in the grammar.  This returns
        variables (mapping from variable name to STIX object ID or
        placeholder), objects (mapping from ID to a STIX object dict), and
        dependencies (mapping from variable name to a set of names of all of
        its dependency variables), as a 3-tuple.
        """
        return self.__variables, self.__objects, self.__dependencies

    def variable_declaration_statement(self, children):
        """
        Updates the variable/object/dependency maps for all variables in the
        statement.
        """
        var_type = children[-1]  # SDO type is always last

        for var_count, var_name, var_props in children[:-1]:

            if var_name in self.__variables:
                raise RedeclaredVariableError(var_name)

            var_dependencies = None

            # Determine if there are any dependency variables
            if var_props:
                var_collector = _VariableDependencyCollector()
                var_collector.visit(var_props)
                var_dependencies = var_collector.variables

            if var_dependencies:
                # If there are dependencies, add a placeholder object to the
                # variables map for now; we need to replace these with actual
                # STIX objects later.  Also store the dependencies for later.
                self.__variables[var_name] = _VariableSpec(
                    var_name, var_count, var_type, var_props
                )
                self.__dependencies[var_name] = var_dependencies

            else:
                # Otherwise, we can directly create an object
                overlay_props = None
                if var_props:
                    builder = _GraphBuilder(
                        self.__object_generator,
                        self.__variables,
                        self.__objects
                    )
                    overlay_props = builder.transform(var_props)

                ids = []
                for _ in range(var_count):
                    obj = _make_object(
                        self.__object_generator, var_type, overlay_props
                    )
                    self.__objects[obj["id"]] = obj
                    ids.append(obj["id"])

                self.__variables[var_name] = ids

        # A formality; no one uses the return value.  But I should probably
        # return something, since all these callbacks are treated as returning
        # a value.
        return None

    def variable_declaration(self, children):
        """
        Produces count/varname/propblock 3-tuples.  The last value (prop block)
        may be None if no property block was given.
        """
        if is_tree(children[0], "count"):
            var_count = int(children[0].children[0])
            var_name = children[1]
        else:
            var_count = 1
            var_name = children[0]

        var_props = None
        if len(children) > 1 and is_tree(children[-1], "property_block"):
            var_props = children[-1]

        # Just forward the property block to the next level up, so we can
        # process the var name, count, type, and prop block together.  (We
        # don't know the variable type at this point, so we can't do anything
        # further here.)
        return var_count, var_name, var_props


class _GraphBuilder(lark.Transformer):
    """
    Selectively processes just the graph statements in a parse tree, and
    produces all of the objects described therein.  No variables are created;
    references to undeclared variables will produce errors.
    """

    def __init__(self, object_generator, variables, objects):
        """
        Initialize this transformer.

        :param object_generator: The object generator to use for making new
            objects.
        :param variables: A mapping from variable name to list of STIX IDs,
            which defines all variables usable in the graph statements.
        :param objects: A mapping from STIX ID to STIX object as dict, which
            contains values for the variables in the variables mapping, and any
            other related objects.  All new objects this transformer creates
            will also be placed in this mapping.
        """
        super().__init__()

        self.__object_generator = object_generator
        self.__variables = variables
        self.__objects = objects

    @lark.v_args(meta=True)
    def graph_statement(self, children, meta):
        """
        Create some objects, optionally connected via a relationship to
        other objects.  Produces the source object IDs only.
        """

        # The below code is simpler if we normalize to a list here, but we
        # still must remember whether this function should propagate a single
        # or list.  So create a new variable instead of modifying source_ids.
        source_ids = children[0]
        if isinstance(source_ids, list):
            source_ids_list = source_ids
        else:
            source_ids_list = [source_ids]

        if len(children) > 1:
            rel_count, rel_type, target_ids = children[1]

            # Same normalization for target IDs.  At least we have the freedom
            # to replace the value of target_ids this time.
            if not isinstance(target_ids, list):
                target_ids = [target_ids]

            for source_id in source_ids_list:

                # Hardcoded special handling of "on": this results in an
                # embedded relationship on the source object whose value is
                # *all* targets, not an SRO per target.  For this reason, it
                # doesn't make sense for the relationship count to be anything
                # but 1.
                if rel_type == "on":
                    if rel_count > 1:
                        raise LanguageError(
                            "Relationship 'on' must have count 1", meta
                        )
                    source_obj = self.__objects[source_id]
                    source_obj["object_refs"] = target_ids

                else:
                    for target_id in target_ids:
                        for _ in range(rel_count):
                            rel = _make_sro(
                                self.__object_generator,
                                source_id,
                                rel_type.value,
                                target_id
                            )
                            self.__objects[rel["id"]] = rel

        return source_ids

    def relationship(self, children):
        """
        Produce a rel_count, rel, target_ids triple representing most of a
        relationship (it doesn't include the source object(s))
        """

        if isinstance(children[0], int):
            rel_count, rel_type, target_ids = children
        else:
            rel_count = 1
            rel_type, target_ids = children

        return rel_count, rel_type, target_ids

    def sdo_list(self, children):
        """
        Produces the concatenation of all STIX object IDs in child nodes
        in a single list.
        """
        all_children = []
        for child in children:
            if isinstance(child, list):
                all_children.extend(child)
            else:
                all_children.append(child)

        return all_children

    def sdo_ref(self, children):
        """
        Produces a single or list of STIX object IDs, either from a
        referenced variable, or from newly created objects.
        """
        if is_token(children[0], "VARIABLE_NAME"):
            # ref is a variable
            var_name = children[0]
            if var_name in self.__variables:
                ids = self.__variables[var_name]
            else:
                raise UndeclaredVariableError(var_name)
        else:
            # ref is to inline SDO(s)
            ids = children[0]

        # A variable or inline SDO with count=1 will produce a single value;
        # count > 1 produces a list.  We need to propagate the right thing,
        # so that STIX properties which expect single IDs don't get lists.
        # (One wishes that such properties would accept a length-1 list, but
        # they don't.)
        if len(ids) == 1:
            ids = ids[0]

        return ids

    def sdo_inline(self, children):
        """Produces a list of the new STIX object IDs."""
        if isinstance(children[0], int):
            count = children[0]
            sdo_type = children[1]
        else:
            count = 1
            sdo_type = children[0]

        overlay_props = None
        if len(children) > 1 and isinstance(children[-1], dict):
            overlay_props = children[-1]

        ids = []
        for _ in range(count):
            obj = _make_object(self.__object_generator, sdo_type, overlay_props)
            self.__objects[obj["id"]] = obj
            ids.append(obj["id"])

        return ids

    def count(self, children):
        """Produces the count, as an int"""
        return int(children[0])

    def property_block(self, children):
        """
        Produces a mapping from prop name to ID(s) or other string literals,
        from the property block.
        """

        overlay_props = {}
        for prop_name, value in children:
            if is_token(value, "ESCAPED_STRING"):
                value = _string_token_value(value)

            overlay_props[prop_name] = value

        return overlay_props

    def property_assignment(self, children):
        """
        Propagates its children.  This will produce a 2-item list including
        the property name and a single or list of STIX IDs which are the "top"
        of the object graph which is the property value.
        """
        return children

    def string_array(self, children):
        """
        Produces a list of strings from the array.
        """
        values = [
            _string_token_value(tok)
            for tok in children
        ]

        return values

    @lark.v_args(meta=True)
    def sighting(self, children, meta):
        """
        Creates a sighting object (and all other related objects).
        """
        sighting_of = overlay_props = None

        if len(children) == 1:
            if isinstance(children[0], dict):
                overlay_props = children[0]
                sighting_of = None
            else:
                overlay_props = None
                sighting_of = children[0]

        elif len(children) > 1:
            overlay_props = children[0]
            sighting_of = children[1]

        if sighting_of and overlay_props and "sighting_of_ref" in overlay_props:
            raise LanguageError(
                "property 'sighting_of_ref' can't both be given explicitly and "
                "in the property block",
                meta
            )

        obj = _make_sighting(
            self.__object_generator, sighting_of, overlay_props
        )
        self.__objects[obj["id"]] = obj

        # A formality; no one uses the return value.  But I should probably
        # return something, since all these callbacks are treated as returning
        # a value.
        return None


class LanguageProcessor:
    """
    Instances process stix prototyping language, and produce stix objects.
    """
    def __init__(self, object_generator, stix_version="2.1"):
        """
        Initialize this processor.

        :param object_generator: The object generator to use
        :param stix_version: Which version of STIX to use.
        :raises stix2generator.exceptions.RegistryNotFoundError: If there isn't
            a built-in registry for the given STIX version
        :raises IOError: (python 2) If the registry JSON file couldn't be opened
            or read
        :raises OSError: (python 3) If the registry JSON file couldn't be opened
            or read (IOError is retained as an alias for backward
            compatibility).
        """
        self.__object_generator = object_generator
        self.__stix_version = stix_version

    def build_graph(
        self, graph_spec, return_variable_bindings=False,
        embed_variable_names=False
    ):
        """
        Build STIX objects from the given specification, expressed with the
        STIX prototyping language.

        :param graph_spec: The graph specification
        :param return_variable_bindings: Whether the caller wants the variable
            bindings returned
        :param embed_variable_names: Whether variable names should be embedded
            in generated objects bound to a variable, using a custom property
        :return: If return_variable_bindings is False, a list of STIX objects.
            Otherwise, return a 2-tuple where the first item is the list of
            generated objects, and the second is a mapping from variable name
            to list of STIX IDs.  This latter value represents the variable
            bindings.
        :raises stix2generator.builder.LanguageError: If there are any errors
            processing the specification itself
        :raises stix2generator.make_object.ObjectGenerationError: If there is an
            error randomly generating a STIX object
        :raises stix2.exceptions.STIXError: If the objects and/or relationships
            expressed in the language, aren't valid STIX
        """
        logger = _get_logger()

        parse_tree = _parser.parse(graph_spec)

        if logger.isEnabledFor(stix2generator.logging.EXTRA_VERBOSE):
            logger.log(stix2generator.logging.EXTRA_VERBOSE, "Parse tree:")
            _print_parse_tree(logger, parse_tree)

        variables, objects, dependencies = \
            _VariableProcessor(self.__object_generator).transform(parse_tree)

        if dependencies:
            self.__process_variable_dependencies(
                variables, objects, dependencies
            )

        if variables and logger.isEnabledFor(logging.DEBUG):
            logger.debug("Variables:")
            for var, ids in variables.items():
                logger.debug("%s = %s", var, ids)

        # This is necessary so that GraphBuilder doesn't re-visit property
        # blocks in variable declarations, and erroneously create extra objects.
        _remove_variable_declaration_statements(parse_tree)

        graph_builder = _GraphBuilder(
            self.__object_generator, variables, objects
        )
        graph_builder.transform(parse_tree)

        # Embed variable names if requested
        if embed_variable_names:
            for var, ids in variables.items():
                for id_ in ids:
                    objects[id_][_VARIABLE_NAME_PROPERTY] = var.value

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Objects:")
            for obj in objects.values():
                logger.debug(pprint.pformat(obj))

        stix_objs = [
            stix2.parse(
                stix_dict, version=self.__stix_version, allow_custom=True
            )
            for stix_dict in objects.values()
        ]

        if return_variable_bindings:
            # The internal variables mapping is keyed by tokens; I don't think
            # that makes sense for users.  So make a new map keyed by plain
            # string variable names.
            variables_plain_names = {
                var.value: ids
                for var, ids in variables.items()
            }
            ret = stix_objs, variables_plain_names
        else:
            ret = stix_objs

        return ret

    def __process_variable_dependencies(self, variables, objects, dependencies):
        """
        Create values for all variables with dependencies.

        :param variables: A mapping with any already-initialized variables, and
            all of the VariableSpec placeholders which were created.  This will
            be updated with newly created variables.
        :param objects: A mapping to store all newly created objects in.
        :param dependencies: The dependency data, as a mapping from var name to
            a set of dependencies.  (This is basically an adjacency list
            representation of a directed graph.)
        :raises UndeclaredVariableError: If there are any references to
            undeclared variables
        :raises CircularVariableDependenciesError: If circular dependencies are
            encountered
        """

        logger = _get_logger()

        sorted_var_names = _topo_sort_dependencies(dependencies)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "Variable creation order: %s",
                ", ".join(tok.value for tok in sorted_var_names)
            )

        for var_name in sorted_var_names:

            if var_name not in variables:
                raise UndeclaredVariableError(var_name)

            var_spec = variables[var_name]

            # Our "leaf node" variables will be regular ones, not
            # VariableSpec's; we don't need to do anything for those.
            if not isinstance(var_spec, _VariableSpec):
                continue

            builder = _GraphBuilder(self.__object_generator, variables, objects)
            overlay_props = builder.transform(var_spec.var_props_tree)

            ids = []
            for _ in range(var_spec.var_count):
                obj = _make_object(
                    self.__object_generator, var_spec.var_type, overlay_props
                )
                ids.append(obj["id"])
                objects[obj["id"]] = obj

            # Now we can replace the spec with real IDs
            variables[var_name] = ids



import distutils.util
import enum
import logging


class Config:
    """
    Simple config class which just holds a set of attributes and values.
    """

    # Subclasses must override this with some actual property names and
    # values.
    _DEFAULTS = {}

    def __init__(self, **kwargs):
        """
        Initialize a config object.  Recognized keywords are those from the
        _DEFAULTS dict.  Any keywords may be omitted, and they will
        automatically be set with default values from the dict.  All valid
        config items are set as attributes on the object, for easy access.

        :param kwargs: User-supplied config attributes and values
        """
        log = logging.getLogger(type(self).__name__)

        for attr_name, default_value in self._DEFAULTS.items():
            in_value = kwargs.get(attr_name, default_value)

            # If given a string but the corresponding default value indicates
            # the config setting is not string-valued, try to convert the given
            # string to the proper type for the config setting.  Otherwise, use
            # the given value as-is.
            if isinstance(in_value, str) and not isinstance(default_value, str):
                attr_type = type(default_value)
                if attr_type == bool:
                    attr_value = bool(distutils.util.strtobool(in_value))
                elif issubclass(attr_type, enum.Enum):
                    # Assumes enum names are all uppercase.
                    attr_value = attr_type[in_value.upper()]
                else:
                    attr_value = attr_type(in_value)

            else:
                attr_value = in_value

            setattr(self, attr_name, attr_value)

        unrecognized_attrs = kwargs.keys() - self._DEFAULTS.keys()
        if unrecognized_attrs:
            log.warning(
                "Unrecognized config attribute(s): %s",
                ", ".join(unrecognized_attrs)
            )



import itertools
import logging
import math
import random
import string

import stix2generator.generation
import stix2generator.generation.constraints
import stix2generator.generation.semantics

from stix2generator.exceptions import (
    CyclicSpecificationReferenceError, InvalidPropertyGroupError,
    ObjectGenerationError, PresenceCoconstraintError,
    SemanticValueTypeMismatchError, SpecificationNotFoundError,
    TypeMismatchError, UndefinedPropertyError, UnrecognizedJSONTypeError,
    ValueCoconstraintError
)


# lazy-initialized
_log = None


_NONE_TYPE = type(None)


# Legal values for the "type" property in specifications
# This and the values of _JSON_TYPE_MAP should of course be
# kept in sync.
_JSON_TYPES = {
    "string",
    "number",
    "integer",
    "boolean",
    "object",
    "array",
    "null"
}


# Used to infer a JSON type from a python type
_JSON_TYPE_MAP = {
    int: "integer",
    float: "number",
    str: "string",
    bool: "boolean",
    _NONE_TYPE: "null",
    dict: "object",
    list: "array"
}


class Config(stix2generator.generation.Config):
    """
    Simple config class for the ObjectGenerator.

    Most properties are self-explanatory.  Some notes:

    - The number settings apply to both "number" and "integer" specs

    - optional_property_probability is the probability of inclusion of
      optional properties.  Must be a number in [0, 1].

    - Setting minimize_ref_properties to True will cause the generator
      to attempt to minimize reference properties in generated objects,
      while still satisfying specification constraints.  If a reference
      property is necessary to satisfy a constraint, it will be included
      regardless of this setting.

      This setting overrides optional_property_probability for reference
      properties: if minimizing reference properties and
      optional_property_probability is 1 (meaning include all optional
      properties), optional reference properties may nevertheless be
      omitted.
    """

    _DEFAULTS = {
        "string_length_min": 5,
        "string_length_max": 20,
        "string_chars": string.ascii_letters,
        "number_min": -1000.0,
        "is_number_min_exclusive": False,
        "number_max": 1000.0,
        "is_number_max_exclusive": False,
        "array_length_min": 1,
        "array_length_max": 5,
        "optional_property_probability": 0.5,
        "minimize_ref_properties": True
    }


class ObjectGenerator:
    """
    Instances of this class generate JSON data from a specification.  More
    specifically, JSON-serializable Python structures are generated, which can
    be converted to JSON text.  So JSON objects are generated as dicts; arrays
    as lists, etc.
    """

    def __init__(
            self, spec_registry=None, semantic_providers=None, config=None
    ):
        """
        Initialize the generator.

        :param spec_registry: A name->specification mapping used to look
            up references inside of specifications
        :param semantic_providers: A list of semantic providers (e.g.
            instances of subclasses of SemanticsProvider)
        :param config: A Config instance giving user settings regarding
            generation.  If None, defaults will be used.
        """
        self.__specs = spec_registry or {}
        self.__config = config or Config()
        self.__semantics = {}

        # Create a map for fast lookup from a semantic name to its provider
        if semantic_providers:
            for provider in semantic_providers:
                semantics = provider.get_semantics()
                for semantic in semantics:
                    self.__semantics[semantic] = provider

    @property
    def config(self):
        """
        Get this generator's config object.
        """
        # If a public property, semantics implementations can use the config to
        # create values compatible with the invoking generator's settings.
        # Might be helpful?
        return self.__config

    @property
    def spec_names(self):
        """
        Get a read-only iterable view of this generator's supported spec
        names.

        :return: An iterable of spec names
        """
        return self.__specs.keys()

    def generate(
        self, spec_name, expected_type=None, spec_name_stack=None,
        value_constraint=None
    ):
        """
        Generate a value based on a specification identified by name.  The name
        is looked up in this generator's registry.

        :param spec_name: The specification name
        :param expected_type: If the named spec should produce a particular
            JSON type, that type.  If it doesn't matter, pass None.  This
            can be used to identify reference errors, where the reference is
            to a specification for the wrong type of thing.
        :param spec_name_stack: A stack of previously-visited specification
            names, used for reference loop detection.  Pass None to start a
            new stack.
        :param value_constraint: A ValueConstraint instance representing some
            additional constraint to be honored by the generator.  This is
            derived from a value co-constraint expression.  If None, there is
            no additional constraint.
        :return: The generated value
        :raises stix2generator.exceptions.CyclicSpecificationReferenceError:
            If a circular reference chain is detected
        :raises stix2generator.exceptions.SpecificationNotFoundError: If the
            named specification is not found in the registry
        :raises stix2generator.exceptions.ObjectGenerationError: For many other
            types of generation errors
        """

        if spec_name_stack is None:
            spec_name_stack = []

        # Reference loop check
        if spec_name in spec_name_stack:
            cycle = spec_name_stack[spec_name_stack.index(spec_name):]
            cycle.append(spec_name)
            raise CyclicSpecificationReferenceError(
                cycle, spec_name_stack=spec_name_stack[:]
            )

        if spec_name not in self.__specs:
            raise SpecificationNotFoundError(
                spec_name, spec_name_stack=spec_name_stack
            )

        spec = self.__specs[spec_name]

        try:

            spec_name_stack.append(spec_name)

            sdo_dict = self.generate_from_spec(
                spec, expected_type=expected_type,
                spec_name_stack=spec_name_stack,
                value_constraint=value_constraint
            )

            spec_name_stack.pop()

        except ObjectGenerationError as e:

            # In a recursive context, set this at the deepest nesting level
            # only
            if not e.spec_name_stack:
                e.spec_name_stack = spec_name_stack[:]  # a copy, to be safe

            raise

        return sdo_dict

    def generate_from_spec(
        self, spec, expected_type=None, spec_name_stack=None,
        value_constraint=None
    ):
        """
        Generate a value based on the given specification, which need not exist
        under any particular name in this generator's registry.

        :param spec: The specification, as parsed JSON
        :param expected_type: If the spec should be for a particular JSON type,
            that type.  If it doesn't matter, pass None.
        :param spec_name_stack: A stack of previously-visited specification
            names, used for reference loop detection.  Pass None to start a
            new stack.
        :param value_constraint: A ValueConstraint instance representing some
            additional constraint to be honored by the generator.  This is
            derived from a value co-constraint expression.  If None, there is
            no additional constraint.
        :return: The generated value
        :raises stix2generator.exceptions.UnrecognizedJSONTypeError: If given a
            non-const dict spec whose declared type is not recognized as a
            JSON type, or if expected_type is given and not a recognized JSON
            type.
        :raises stix2generator.exceptions.TypeMismatchError: If expected_type is
            given and the spec type doesn't match.
        :raises stix2generator.exceptions.ObjectGenerationError: For various
            ways the given spec is invalid.  Other types of errors are also
            wrapped/chained from this exception type (if possible) so that
            we get decoration with extra info from higher stack frames, which
            is useful for diagnosing where those problems occur.
        """

        spec_type = _get_spec_type(spec)

        if expected_type:
            if expected_type not in _JSON_TYPES:
                raise UnrecognizedJSONTypeError(expected_type)

            # There really should be some flexibility for numeric types: if
            # number is expected, integers should be accepted too...
            if spec_type != expected_type:
                raise TypeMismatchError(
                    expected_type, spec_type
                )

        # If not a dict, the spec IS the desired value.  It's an easy way to
        # produce fixed values.
        if not isinstance(spec, dict):
            value = spec

        # The other way: use "const", like in json-schema.
        elif "const" in spec:
            value = spec["const"]

        else:

            semantic_name = spec.get(
                stix2generator.generation.semantics.SEMANTIC_PROPERTY_NAME
            )

            try:

                if semantic_name:
                    value = self.__generate_semantic(spec, value_constraint)

                else:
                    value = self.__generate_plain(
                        spec, spec_name_stack, value_constraint
                    )

            except ObjectGenerationError as e:
                # In a recursive context, set this at the deepest nesting
                # level only.  Also, I think it's better to use the semantic
                # name as the type name in error messages, for semantic specs.
                if not e.spec_type:
                    e.spec_type = semantic_name or spec_type
                raise

            except Exception as e:
                raise ObjectGenerationError(
                    "An error occurred during generation: {}: {}".format(
                        type(e).__name__, str(e)
                    ),
                    semantic_name or spec_type
                ) from e

        return value

    def __generate_semantic(self, spec, value_constraint):
        """
        Generate from a semantic-type spec.

        :param spec: The spec
        :param value_constraint: A ValueConstraint instance representing some
            additional constraint to be honored by the generator.  This is
            derived from a value co-constraint expression.  If None, there is
            no additional constraint.
        :return: The generated value
        :raises stix2generator.exceptions.SemanticValueTypeMismatchError: If the
            semantic produces a value which doesn't agree with the spec's
            declared type.
        :raises stix2generator.exceptions.ObjectGenerationError: If the
            semantic name isn't found in any of this generator's semantic
            providers
        """
        semantic = spec[
            stix2generator.generation.semantics.SEMANTIC_PROPERTY_NAME
        ]

        if semantic in self.__semantics:
            provider = self.__semantics[semantic]

            value = provider.create_semantic(spec, self, value_constraint)

            # Should check that the implementation created the right type of
            # value.
            actual_type = _json_type_from_python_type(type(value))

            if actual_type != spec["type"]:
                raise SemanticValueTypeMismatchError(
                    semantic, actual_type, value, spec["type"]
                )

        else:
            raise ObjectGenerationError(
                "unrecognized semantic: " + semantic
            )

        return value

    def __generate_plain(self, spec, spec_name_stack, value_constraint):
        """
        Generate from a "plain" spec, i.e. one that isn't a semantic spec.
        Maybe not a very good method name... (Of course, the spec can have
        sub-specs or references to specs which are semantic.)

        :param spec: The "plain" spec
        :param spec_name_stack: Spec name stack, to propagate through
            references (if any)
        :param value_constraint: A ValueConstraint instance representing some
            additional constraint to be honored by the generator.  This is
            derived from a value co-constraint expression.  If None, there is
            no additional constraint.  This is *only* used to propagate through
            ref/oneOf specifications.  The "plain" generators (non-semantic)
            ignore constraints.
        :return: The generated value
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """
        type_ = spec["type"]

        if "ref" in spec:
            value = self.generate(
                spec["ref"],
                expected_type=type_,
                spec_name_stack=spec_name_stack,
                value_constraint=value_constraint
            )

        elif "oneOf" in spec:
            # value of the "oneOf" property should be a list of specs.
            sub_spec = random.choice(spec["oneOf"])
            value = self.generate_from_spec(
                sub_spec,
                expected_type=type_,
                spec_name_stack=spec_name_stack,
                value_constraint=value_constraint
            )

        else:
            generator = self.__GENERATOR_METHOD_MAP[type_]
            value = generator(self, spec, spec_name_stack)

        return value

    def generate_object(self, object_spec, spec_name_stack=None):
        """
        Generate a JSON object from the given specification.

        :param object_spec: A JSON object specification
        :param spec_name_stack: Specification name stack, for reference loop
            detection.  If None, use an empty stack.
        :return: A dict
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """

        # Handle imports
        if "import" in object_spec:
            imported_spec_name = object_spec["import"]
            gen_object = self.generate(
                imported_spec_name,
                expected_type="object",
                spec_name_stack=spec_name_stack
            )
        else:
            gen_object = {}

        # First, determine which properties to include
        names_to_include = _get_properties_to_include(
            object_spec, self.config.optional_property_probability,
            self.config.minimize_ref_properties
        )

        if names_to_include:
            # Then, find values for the included properties, according to their
            # specs and co-constraints.
            value_coconstraints = _get_value_coconstraints(object_spec)

            # At this point, if there were any names to include, there must
            # have been some properties defined!
            prop_specs = object_spec["properties"]
            for prop_name in names_to_include:

                # Generate constraint if necessary
                constraint = _get_value_constraint(
                    prop_name, value_coconstraints, gen_object
                )

                gen_object[prop_name] = self.generate_from_spec(
                    prop_specs[prop_name], spec_name_stack=spec_name_stack,
                    value_constraint=constraint
                )

        return gen_object

    def generate_array(self, array_spec, spec_name_stack=None):
        """
        Generate a JSON array from the given specification.

        :param array_spec: A JSON array specification
        :param spec_name_stack: Specification name stack, for reference loop
            detection.  If None, use an empty stack.
        :return: A list
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """

        item_spec = array_spec["items"]

        has_min = "minItems" in array_spec
        has_max = "maxItems" in array_spec

        if (has_min and not has_max) or (not has_min and has_max):
            raise ObjectGenerationError(
                "Specification must include both or neither of the properties: "
                "minItems, maxItems",
                "array"
            )

        min_items = array_spec.get("minItems", self.config.array_length_min)
        max_items = array_spec.get("maxItems", self.config.array_length_max)

        if min_items > max_items:
            raise ObjectGenerationError(
                "minItems must be less than or equal to maxItems",
                "array"
            )

        if min_items < 0 or max_items < 0:
            raise ObjectGenerationError(
                "minItems and maxItems must be non-negative: {}".format(
                    min_items if min_items < 0 else max_items
                ),
                "array"
            )

        array = [
            self.generate_from_spec(item_spec, spec_name_stack=spec_name_stack)
            for _ in range(
                random.randint(min_items, max_items)
            )
        ]

        return array

    def generate_string(self, string_spec, spec_name_stack=None):
        """
        Generate a string from the given specification.

        :param string_spec: A string specification
        :param spec_name_stack: A specification name stack, for reference loop
            detection.  Unused but included for API compatibility with
            object/array generators.
        :return: A string
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """

        has_min = "minLength" in string_spec
        has_max = "maxLength" in string_spec

        if (has_min and not has_max) or (not has_min and has_max):
            raise ObjectGenerationError(
                "Specification must include both or neither of the properties: "
                "minLength, maxLength",
                "string"
            )

        min_length = string_spec.get("minLength", self.config.string_length_min)
        max_length = string_spec.get("maxLength", self.config.string_length_max)

        if min_length > max_length:
            raise ObjectGenerationError(
                "minLength must be less than or equal to maxLength: {} <= {}"
                .format(min_length, max_length),
                "string"
            )

        if min_length < 0 or max_length < 0:
            raise ObjectGenerationError(
                "minLength and maxLength must be non-negative: {}".format(
                    min_length if min_length < 0 else max_length
                ),
                "string"
            )

        s = "".join(
            random.choice(self.config.string_chars)
            for _ in range(
                random.randint(min_length, max_length)
            )
        )

        return s

    def generate_integer(self, integer_spec, spec_name_stack=None):
        """
        Generate an integer from the given specification.

        :param integer_spec: An integer specification
        :param spec_name_stack: A specification name stack, for reference loop
            detection.  Unused but included for API compatibility with
            object/array generators.
        :return: An int
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """
        min_, is_min_exclusive, max_, is_max_exclusive = \
            _process_numeric_min_max_properties(
                integer_spec,
                self.config.number_min,
                self.config.is_number_min_exclusive,
                self.config.number_max,
                self.config.is_number_max_exclusive
            )

        # Guess I won't assume the user expressed the bounds as ints, so I
        # need to convert to ints and check the resulting bounds.  The
        # call above to process min/max properties doesn't assume we require
        # ints.
        if int(min_) == min_:
            min_ = int(min_)
            if is_min_exclusive:
                min_ += 1
        else:
            min_ = int(math.ceil(min_))

        if int(max_) == max_:
            max_ = int(max_)
            if is_max_exclusive:
                max_ -= 1
        else:
            max_ = int(math.floor(max_))

        if min_ > max_:
            raise ObjectGenerationError(
                "no integers exist in the specified interval",
                "integer"
            )

        return random.randint(min_, max_)

    def generate_number(self, number_spec, spec_name_stack=None):
        """
        Generate a number (float) from the given specification.

        :param number_spec: A number specification
        :param spec_name_stack: A specification name stack, for reference loop
            detection.  Unused but included for API compatibility with
            object/array generators.
        :return: A float
        :raises stix2generator.exceptions.ObjectGenerationError: If a
            generation error occurs
        """

        min_, is_min_exclusive, max_, is_max_exclusive = \
            _process_numeric_min_max_properties(
                number_spec,
                self.config.number_min,
                self.config.is_number_min_exclusive,
                self.config.number_max,
                self.config.is_number_max_exclusive
            )

        if is_min_exclusive and is_max_exclusive:
            n = _random_open(min_, max_)
        elif is_min_exclusive:
            n = _random_half_open_lower(min_, max_)
        elif is_max_exclusive:
            n = _random_half_open_upper(min_, max_)
        else:
            n = _random_closed(min_, max_)

        return n

    def generate_boolean(self, boolean_spec, spec_name_stack=None):
        """
        Generate a boolean from the given specification.

        :param boolean_spec: A boolean specification (ignored; there's nothing
            to configure for now)
        :param spec_name_stack: A specification name stack, for reference loop
            detection.  Unused but included for API compatibility with
            object/array generators.
        :return: True or False
        """
        if random.random() < 0.5:
            return True
        return False

    def generate_null(self, null_spec, spec_name_stack=None):
        """
        Generate null (None).

        :param null_spec: A null specification (ignored; there's nothing
            to configure for now)
        :param spec_name_stack: A specification name stack, for reference loop
            detection.  Unused but included for API compatibility with
            object/array generators.
        :return: None
        """
        return None

    # This has to be at the bottom, after the methods are defined.  Would it
    # have been better to store method names and use getattr() to get the
    # methods instead?  Or generate a function name from a template?  This is
    # yet another map to keep sync'd up with others.  Think about ways of
    # improving this situation...
    __GENERATOR_METHOD_MAP = {
        "object": generate_object,
        "array": generate_array,
        "string": generate_string,
        "integer": generate_integer,
        "number": generate_number,
        "boolean": generate_boolean,
        "null": generate_null
    }


def _get_logger():
    global _log
    if _log is None:
        _log = logging.getLogger(__name__)

    return _log


def _get_value_coconstraints(object_spec):
    """
    Get the value coconstraints, if any, from the given object specification.
    This also does some error checking.

    :param object_spec: The object specification whose value coconstraints
        should be checked.
    :return: A list of ValueCoconstraint objects; will be empty if there are
        none defined.
    :raises stix2generator.exceptions.ValueCoconstraintError: If an
        invalid value co-constraint is found.
    """

    # This function shouldn't be called if object_spec has no properties, but
    # just in case...
    assert "properties" in object_spec

    value_coconstraints = object_spec.get("value-coconstraints", [])
    prop_specs = object_spec["properties"]
    coconstraint_objs = []

    for coconstraint in value_coconstraints:
        coconstraint_obj = \
            stix2generator.generation.constraints.make_value_coconstraint(
                coconstraint
            )

        coconstraint_objs.append(coconstraint_obj)

        if coconstraint_obj.prop_name_left not in prop_specs:
            raise ValueCoconstraintError(
                coconstraint,
                "Property '{}' undefined in specification".format(
                    coconstraint_obj.prop_name_left
                )
            )

        if coconstraint_obj.prop_name_right not in prop_specs:
            raise ValueCoconstraintError(
                coconstraint,
                "Property '{}' undefined in specification".format(
                    coconstraint_obj.prop_name_right
                )
            )

    # Another scan through the coconstraints to check for properties
    # referenced more than once.
    prop_occurrence_counts = {}

    def inc_count_for_key(d, k):
        count = d.setdefault(k, 0) + 1
        d[k] = count

    for coconstraint_obj in coconstraint_objs:
        inc_count_for_key(
            prop_occurrence_counts,
            coconstraint_obj.prop_name_left
        )

        inc_count_for_key(
            prop_occurrence_counts,
            coconstraint_obj.prop_name_right
        )

    props_with_count_gt_1 = [
        k for k, v in prop_occurrence_counts.items()
        if v > 1
    ]

    if props_with_count_gt_1:
        log = _get_logger()
        log.warning(
            "Some properties are referenced in more than one value"
            " co-constraint.  If such a property requires constraining based on"
            " another property value, only the first such co-constraint will be"
            " consulted: %s", ", ".join(props_with_count_gt_1)
        )

    return coconstraint_objs


def _check_property_groups(groups_spec, property_specs):
    """
    Do some sanity checks on the given property groups: empty groups,
    bad property names, naming conflicts, etc.  Runs for side-effects
    (exceptions) and doesn't return anything.

    :param groups_spec: The groups spec from a presence coconstraint
        specification from an object specification
    :param property_specs: The properties specifications from an object
        specification
    :raises stix2generator.exceptions.InvalidPropertyGroupError: If
        there is a problem with a property group
    """
    for group_name, prop_names in groups_spec.items():

        if not prop_names:
            raise InvalidPropertyGroupError(
               group_name, "group can't be empty"
            )

        if group_name in property_specs:
            raise InvalidPropertyGroupError(
                group_name, "group name conflicts with a property name"
            )

        undef_props = set(prop_names) - property_specs.keys()
        if undef_props:
            raise InvalidPropertyGroupError(
                group_name, 'undefined property(s): {}'.format(
                    ", ".join(undef_props)
                )
            )

    # check pairwise intersections to ensure all groups are disjoint
    if len(groups_spec) > 1:
        for group1, group2 in itertools.combinations(groups_spec, 2):
            overlaps = set(groups_spec[group1]) & set(groups_spec[group2])
            if overlaps:
                raise InvalidPropertyGroupError(
                    group2,
                    'overlaps with group "{}".  Overlapping properties: {}'
                    .format(
                        group1,
                        ", ".join(overlaps)
                    )
                )


def _get_group_coconstraints(presence_coconstraints, property_specs):
    """
    Get property group co-constraints from the presence co-constraint
    part of an object specification.

    Length 1 property groups will have some sanity checking done, but
    will otherwise be ignored.  That is better accomplished by using the
    property directly instead of putting it in its own group.

    :param presence_coconstraints: The presence co-constraints dict from an
        object specification.
    :param property_specs: The property definitions dict from an object
        specification.
    :return: A mapping of group name to
        stix2generator.generation.constraints.PresenceCoconstraint object
        representing the group co-constraint.  If there were no property
        groups defined, the map will be empty.
    :raises stix2generator.exceptions.PresenceCoconstraintError: If an invalid
        presence co-constraint is found
    """

    # Should map group names to lists of property names (group contents)
    group_specs = presence_coconstraints.get("property-groups", {})

    # Sanity check groups
    _check_property_groups(group_specs, property_specs)

    all_of_groups = set(presence_coconstraints.get("all", []))
    one_of_groups = set(presence_coconstraints.get("one", []))
    at_least_one_of_groups = set(
        presence_coconstraints.get("at-least-one", [])
    )

    # Sanity check for bad group names in constraint type lists
    for group_name in itertools.chain(
        all_of_groups, one_of_groups, at_least_one_of_groups
    ):
        if group_name not in group_specs:
            raise PresenceCoconstraintError(
                "Group not found: " + group_name
            )

    # Ensure no group is assigned more than one constraint type
    constraint_conflicts = set()
    for groups1, groups2 in itertools.combinations((
        all_of_groups, one_of_groups, at_least_one_of_groups
    ), 2):
        constraint_conflicts |= groups1 & groups2

    if constraint_conflicts:
        raise PresenceCoconstraintError(
            "Property group(s) have conflicting co-constraints: {}".format(
                ", ".join(constraint_conflicts)
            )
        )

    # Define a default constraint type, or require that every group be
    # explicitly assigned one?  Maybe being explicit is clearer?  So check for
    # groups which weren't assigned constraint types.
    unassigned_groups = group_specs.keys() - all_of_groups - \
        one_of_groups - at_least_one_of_groups
    if unassigned_groups:
        raise PresenceCoconstraintError(
            "Property group(s) were not assigned co-constraint types: "
            "{}".format(
                ", ".join(unassigned_groups)
            )
        )

    # Filter out length-1 groups.  I think I'd like the above checks to be done
    # on them anyway, to avoid silly specifications, but here we will start
    # ignoring them.
    groups_to_ignore = []
    for group_name, property_names in group_specs.items():
        if len(property_names) == 1:
            groups_to_ignore.append(group_name)

    group_specs = dict(group_specs)  # shallow copy ok
    for group_name in groups_to_ignore:
        del group_specs[group_name]

    constraint_objs = {
        group_name: stix2generator.generation.constraints.PresenceCoconstraint(
            property_names,
            "one" if group_name in one_of_groups
            else "all" if group_name in all_of_groups
            else "at-least-one"
        )
        for group_name, property_names in group_specs.items()
    }

    return constraint_objs


def _get_dependency_coconstraints(presence_coconstraints, group_coconstraints,
                                  property_specs):
    """
    Get dependency co-constraints, which is part of the presence co-constraints.
    This just does a lot of sanity checks on the dependencies object from the
    spec.

    :param presence_coconstraints:
    :param group_coconstraints:
    :param property_specs:
    :return: The dependency co-constraints object.  If none was given, returns
        an empty dict.
    :raises stix2generator.exceptions.PresenceCoconstraintError: If an invalid
        presence co-constraint is found
    """

    deps = presence_coconstraints.get("dependencies", {})

    grouped_property_names = set(
        itertools.chain.from_iterable(
            coco.property_names
            for coco in group_coconstraints.values()
        )
    )

    def is_group_or_prop(name):
        return name in property_specs or name in group_coconstraints

    for key_name, prop_list in deps.items():
        if not is_group_or_prop(key_name):
            raise PresenceCoconstraintError(
                "Unrecognized group or property: " + key_name
            )

        if key_name in grouped_property_names:
            raise PresenceCoconstraintError(
                'Property "{}" is grouped and cannot be referenced'
                ' individually'.format(key_name)
            )

        for name in prop_list:

            if not is_group_or_prop(name):
                raise PresenceCoconstraintError(
                    "Unrecognized group or property: " + name
                )

            if name in grouped_property_names:
                raise PresenceCoconstraintError(
                    'Property "{}" is grouped and cannot be referenced'
                    ' individually'.format(name)
                )

            if name in deps:
                raise PresenceCoconstraintError(
                    "Dependency key can't also occur in a dependency"
                    " value: " + name
                )

    return deps


def _get_presence_coconstraints(object_spec):
    """
    Get presence co-constraint info from the given object specification.
    This includes the groups and dependencies.

    :param object_spec: The object specification
    :return: A 2-tuple with (a) the group co-constraint mapping from group
        name to constraint object, and the dependencies mapping from property
        or group name to list of properties/groups.
    :raises stix2generator.exceptions.PresenceCoconstraintError: If an invalid
        presence co-constraint is found
    """

    presence_coconstraints = object_spec.get("presence-coconstraints", {})
    property_specs = object_spec.get("properties", {})

    group_coconstraints = _get_group_coconstraints(
        presence_coconstraints, property_specs
    )

    dependency_coconstraints = _get_dependency_coconstraints(
        presence_coconstraints, group_coconstraints, property_specs
    )

    return group_coconstraints, dependency_coconstraints


def _get_properties_to_include(
        object_spec, optional_property_probability, minimize_ref_properties
):
    """
    Determine which object properties to include, based on required/optional
    choices and any defined presence co-constraints.

    :param object_spec: The object spec
    :param optional_property_probability: The probability an optional property
        should be included.  Must be a number from 0 to 1.
    :param minimize_ref_properties: True if we should minimize optional
        reference properties.  False if they should receive no special
        treatment.
    :return: The property names, as a set of strings
    :raises stix2generator.exceptions.PresenceCoconstraintError: If an invalid
        presence co-constraint is found
    :raises stix2generator.exceptions.UndefinedPropertyError: If a reference to an
        undefined property or group is found in the "required" or "optional"
        property value of the spec
    :raises stix2generator.exceptions.ObjectGenerationError: If a reference to a
        grouped property is found
    """

    prop_specs = object_spec.get("properties", {})
    required_names = object_spec.get("required")
    optional_names = object_spec.get("optional")

    if required_names is not None and optional_names is not None:
        raise ObjectGenerationError(
            '"required" and "optional" can\'t both be present'
        )

    # If neither optional nor required names are specified, all
    # properties/groups will be required.
    elif required_names is None and optional_names is None:
        # empty optional set = all required
        optional_names = set()

    # Convert to sets to remove dupes
    elif required_names is not None:
        required_names = set(required_names)
    elif optional_names is not None:
        optional_names = set(optional_names)

    group_coconstraints, dependency_coconstraints = \
        _get_presence_coconstraints(object_spec)

    # Detect errors in the required/optional prop list: all must be
    # defined, and grouped properties must not be referenced
    req_or_opt = required_names if required_names is not None \
        else optional_names
    defined_prop_names = prop_specs.keys()
    defined_group_names = group_coconstraints.keys()
    grouped_property_names = set(
        itertools.chain.from_iterable(
            coco.property_names
            for coco in group_coconstraints.values()
        )
    )

    undef_name_errors = req_or_opt - defined_prop_names - defined_group_names
    if undef_name_errors:
        raise UndefinedPropertyError(undef_name_errors)

    grouped_prop_errors = req_or_opt & grouped_property_names
    if grouped_prop_errors:
        raise ObjectGenerationError(
            "Property(s) are grouped and cannot be referenced"
            " individually: {}".format(
                ", ".join(
                    "{}".format(p) for p in grouped_prop_errors
                )
            )
        )

    # Include all ungrouped property names and property group names in
    # the same "pool" of names one can specify as required or optional.
    name_pool = (defined_prop_names - grouped_property_names) \
        | defined_group_names

    # Get set of optional names (whether they specified "required" or
    # "optional" in the spec).
    effectively_optional_names = optional_names if optional_names is not None \
        else name_pool - required_names

    # Start out the set of names to include with all required ones.
    names_to_include = required_names if required_names is not None \
        else name_pool - effectively_optional_names

    # And then maybe add some optional ones.
    for name in effectively_optional_names:

        is_group = name in defined_group_names
        is_ref = name.endswith("_ref") or name.endswith("_refs")

        can_include = False
        if minimize_ref_properties:
            if is_group:
                if group_coconstraints[name].can_satisfy_without_refs():
                    can_include = True
            elif not is_ref:
                can_include = True

        else:
            can_include = True

        if can_include and random.random() < optional_property_probability:
            names_to_include.add(name)

    # Incorporate the "dependencies": add any other properties we
    # require
    for dep_key, dep_names in dependency_coconstraints.items():
        if dep_key in names_to_include:
            names_to_include.update(dep_names)

    # For any names which are property groups, expand them to the
    # component properties according to their co-constraints
    # ... can't modify a set as you iterate!  So need a temp set.
    temp_set = set()
    for name in names_to_include:
        if name in group_coconstraints:
            temp_set.update(
                group_coconstraints[name].choose_properties(
                    optional_property_probability,
                    minimize_ref_properties
                )
            )
        else:
            temp_set.add(name)

    names_to_include = temp_set

    return names_to_include


def _get_value_constraint(prop_name, coconstraints, partially_generated_object):
    """
    Get a value constraint object derived from a value co-constraint involving
    the given property.

    :param prop_name: The property to check for co-constraints
    :param coconstraints: An iterable of co-constraint objects
        (stix2generator.generation.constraints.ValueCoconstraint)
    :param partially_generated_object: The object being generated, in its
        current state of partial construction.  This is necessary to find the
        value of the other property involved in the co-constraint, if any.
    :return: A constraint object
        (stix2generator.generation.constraints.ValueConstraint) representing a
        necessary constraint on the given property, or None if no constraint is
        necessary.
    """
    constraint = None
    for coconstraint in coconstraints:
        if coconstraint.involves_property(prop_name):
            other_prop_name = coconstraint.get_other_property(
                prop_name
            )
            if other_prop_name in partially_generated_object:
                constraint = coconstraint.get_constraint(
                    other_prop_name,
                    partially_generated_object[other_prop_name]
                )

            break

    return constraint


def _json_type_from_python_type(python_type):
    """
    Infers a JSON type from a python type.  This is necessary for reference
    verification when the type isn't explicitly given (e.g. a "const" spec).

    :param python_type: The python type (a 'type' object)
    :return: A JSON type name
    :raises stix2generator.exceptions.ObjectGenerationError: If a JSON type can't
        be inferred from python_type
    """
    json_type = _JSON_TYPE_MAP.get(python_type)
    if json_type is None:
        raise ObjectGenerationError(
            "Can't infer JSON type from " + str(python_type)
        )

    return json_type


def _get_spec_type(spec):
    """
    Determine the type of the given spec, as one of the supported JSON
    types.

    :param spec: A specification
    :return: A spec JSON type as a string, e.g. "string", "array", etc.
    :raises stix2generator.exceptions.ObjectGenerationError: If a const spec
        where the spec type can't be inferred from the constant; if a
        non-const spec whose "type" property is missing
    :raises stix2generator.exceptions.UnrecognizedJSONTypeError: if the value of
        the "type" property isn't a recognized JSON type
    """
    if isinstance(spec, dict):
        if "const" in spec:
            # type is implied by the value of the "const" property.
            value_type = type(spec["const"])
            json_type = _json_type_from_python_type(value_type)

        elif "type" in spec:
            json_type = spec["type"]
            if json_type not in _JSON_TYPES:
                raise UnrecognizedJSONTypeError(json_type)

        else:
            raise ObjectGenerationError(
                '"type" property is missing'
            )

    else:
        # the spec is the value.  Check its type.
        value_type = type(spec)
        json_type = _json_type_from_python_type(value_type)

    return json_type


def _process_numeric_min_max_properties(
    spec,
    default_min,
    is_default_min_exclusive,
    default_max,
    is_default_max_exclusive
):
    """
    Factors out a rather large chunk of code for validating and processing
    the min/max properties on numbers and integers.  Maybe we need a
    JSON-Schema for specifications and validate against that, to reduce the
    amount of hand-written validation code we need to write...

    :param spec: A number or integer spec
    :param default_min: If the spec doesn't specify a minimum, use this as
        the default.
    :param is_default_min_exclusive: Whether default_min, if it is used,
        is an exclusive bound.
    :param default_max: If the spec doesn't specify a maximum, use this as
        the default.
    :param is_default_max_exclusive: Whether default_max, if it is used,
        is an exclusive bound.
    :return: A (num, bool, num, bool) 4-tuple giving the bounds and whether
        each bound is exclusive or not:
            (min, is_min_exclusive, max, is_max_exclusive)
    :raises stix2generator.exceptions.ObjectGenerationError: For various types
        of problems with numeric specifications
    """
    if "minimum" in spec and "exclusiveMinimum" in spec:
        raise ObjectGenerationError(
            "minimum and exclusiveMinimum can't both be present"
        )

    if "maximum" in spec and "exclusiveMaximum" in spec:
        raise ObjectGenerationError(
            "maximum and exclusiveMaximum can't both be present"
        )

    min_given = any(
        p in spec
        for p in ("minimum", "exclusiveMinimum")
    )
    max_given = any(
        p in spec
        for p in ("maximum", "exclusiveMaximum")
    )

    # I think this check is necessary since user-specified min/max could well
    # be out of order w.r.t. defaults, producing unexpected errors.  What would
    # users expect the other bound to be anyway, if they only gave one bound?
    if (min_given and not max_given) or (max_given and not min_given):
        raise ObjectGenerationError(
            "can't give minimum without a maximum, or vice versa"
        )

    if "minimum" in spec:
        min_ = spec["minimum"]
        is_min_exclusive = False
    elif "exclusiveMinimum" in spec:
        min_ = spec["exclusiveMinimum"]
        is_min_exclusive = True
    else:
        min_ = default_min
        is_min_exclusive = is_default_min_exclusive

    if "maximum" in spec:
        max_ = spec["maximum"]
        is_max_exclusive = False
    elif "exclusiveMaximum" in spec:
        max_ = spec["exclusiveMaximum"]
        is_max_exclusive = True
    else:
        max_ = default_max
        is_max_exclusive = is_default_max_exclusive

    if min_ > max_:
        raise ObjectGenerationError(
            "minimum can't be greater than maximum"
        )
    elif min_ == max_ and (is_max_exclusive or is_min_exclusive):
        raise ObjectGenerationError(
            "In an open or half-open interval, minimum must be strictly "
            "less than maximum"
        )

    return min_, is_min_exclusive, max_, is_max_exclusive


def _random_half_open_upper(min_, ex_max):
    assert min_ < ex_max

    # easy case... I think, since random.random() already has the right
    # openness.
    n = min_ + (ex_max - min_) * random.random()

    return n


def _random_half_open_lower(ex_min, max_):
    assert ex_min < max_

    # harder case: we compute the opposite openness, then "flip" it by
    # negating the result.  So for example, [a,b) becomes (-b,-a].  That
    # gives us the correct openness and range, but wrong endpoints.  Then, we
    # just "shift" the interval to its proper endpoints.  That's one way of
    # looking at it, at least.
    n = ex_min + max_ - _random_half_open_upper(ex_min, max_)

    return n


def _random_closed(min_, max_):
    assert min_ <= max_

    # Python gives us a simple API which is documented to use a closed
    # interval, but the same docs say that one endpoint may or may not actually
    # be included... so maybe this doesn't actually work?  It actually quotes
    # the exact same equation I used in _random_half_open_upper()!
    return random.uniform(min_, max_)


def _random_open(ex_min, ex_max):
    assert ex_min < ex_max

    # It's not obvious how to have a uniformly distributed open interval.  I
    # had an idea to add two intervals of opposite openness covering the same
    # range and divide by 2, to obtain a totally open interval.  E.g. to get
    # (0, 1) compute ([0, 1) + (0, 1]) / 2.  I think this yields an open
    # interval, but at the expense of uniformity.
    # n = _random_half_open_upper(ex_min, ex_max) + \
    #     _random_half_open_lower(ex_min, ex_max)
    # n /= 2.0

    # Another idea is to split an open interval into two half-open intervals
    # which are joined at a closed boundary in the middle.  This implies that
    # the mid point is slightly more likely than other points, so this isn't
    # uniform either, but I think it's better than above.  Perhaps it's close
    # enough??
    mid = (ex_min + ex_max) / 2.0
    if random.random() < 0.5:
        n = _random_half_open_lower(ex_min, mid)
    else:
        n = _random_half_open_upper(mid, ex_max)

    return n

    # Another possibility I found on stackoverflow was essentially to generate
    # a random positive integer (1 to something big, to give a lot of possible
    # distinct generated numbers) and divide by a number slightly larger than
    # the maximum.  https://stackoverflow.com/a/19934205



import collections
import collections.abc
import enum
import random
import stix2.utils

import stix2generator
import stix2generator.utils
import stix2generator.generation
import stix2generator.generation.object_generator


class InversePropertyConstraint:
    """
    Instances represent an "inverse" property constraint.  This type of
    constraint means that reference properties in each of two objects are
    "inverses" with respect to each other: if you set one to refer to a
    particular object, the inverse property of the other object must be set to
    refer to the first object.  They *must* refer to each others' objects.  An
    example of this is directory/content_refs and file/parent_directory_ref.
    This necessarily implies cycles in the graph.

    Constraint specification is simple: you just give two STIX object types and
    the property name from each which are inverses with respect to each other.
    This specification is also symmetric: (type1, A, type2, B) means the same
    as (type2, B, type1, A), where A and B are property names.

    Application of the constraint to a pair of objects however, may not be so
    simple.  The type and property names need not be distinct.  This can result
    in ambiguity in applying the constraint.  When type names are equal, an
    instance of the type may have both constraint properties.  Then, the
    constraint can apply in two different ways.  For example, consider the
    process SCO, which has child_refs and parent_ref properties, which are
    inverses:

        process1:
            child_refs=[2]
            parent_ref=2

        process2:
            child_refs=[3]
            parent_ref=1

    process1's child_refs refers to process2, so the constraint says process2's
    parent_ref must refer to process1.  It does, so the constraint holds in
    that case.  But process1's parent_ref refers to process2, and process2's
    child_refs does not refer to process1.  So the constraint does not hold in
    that case.  Both objects have both constraint properties, so the constraint
    effectively applies twice, and in each case independently, it may or may
    not hold.  So it's not sufficient to ask whether an arbitrary inverse
    property constraint holds between two arbitrary objects.  One must somehow
    express *how* to apply the constraint, in cases where there is ambiguity.

    So method signatures for this class include the ref_prop parameter for this
    purpose.  It expresses which constraint property in the first object
    (src_obj) you are interested in.  In results in asymmetry in the
    application of the constraint, despite the symmetry of the constraint
    itself.  For example, the holds() method is essentially checking for an
    inverse property cycle between two objects.  With respect to the
    (type1, A, type2, B) example above, invoking as holds(src_obj, A, dest_obj)
    means checking for the cycle:

        src_obj -> A -> dest_obj -> B -> src_obj

    whereas invoking as holds(src_obj, B, dest_obj) means checking for the
    cycle:

        src_obj -> B -> dest_obj -> A -> src_obj

    and those cycles are different (assuming A and B are different properties).
    """
    def __init__(self, object_type1, prop_name1, object_type2, prop_name2):
        """
        Initialize this constraint object.  The constraint is simply described
        with two object types and two bare property names.  So you can't
        distinguish same named properties in different locations inside of
        objects.  Only top-level properties are supported.  If constraints need
        to target non-top-level properties, more complexity will be necessary.

        :param object_type1: A STIX object type
        :param prop_name1: A reference property name from an object of type
            object_type1
        :param object_type2: Another STIX object type
        :param prop_name2: A reference property name from an object of type
            object_type2
        """
        self.__object_type1 = object_type1
        self.__prop_name1 = prop_name1
        self.__object_type2 = object_type2
        self.__prop_name2 = prop_name2

    def __repr__(self):
        return 'InversePropertyConstraint("{}", "{}", "{}", "{}")'.format(
            self.object_type1, self.prop_name1,
            self.object_type2, self.prop_name2
        )

    @property
    def object_type1(self):
        return self.__object_type1

    @property
    def prop_name1(self):
        return self.__prop_name1

    @property
    def object_type2(self):
        return self.__object_type2

    @property
    def prop_name2(self):
        return self.__prop_name2

    def is_applicable(self, src_obj, ref_prop, dest_obj):
        """
        Given a source/dest object and a ref property name, determine whether
        this constraint would constrain an inverse property of dest_obj.  To be
        applicable, the source object type and ref_prop must match with one of
        the type/property pairs of this constraint, dest_obj's type must
        match with the other type, it must possess the inverse property, and
        the source object must actually refer to dest_obj via ref_prop.
        dest_obj need not refer back to src_obj via the inverse property (if
        this constraint were actually enforced, it would be).

        :param src_obj: An object
        :param ref_prop: A property name
        :param dest_obj: Another object
        :return: True if this constraint matches the props/types of the objects
            and would constrain a property of dest_obj; False if not.
        """

        src_type = src_obj["type"]
        dest_type = dest_obj["type"]

        # check src/ref_prop matches with object/prop1,
        # dest matches object/prop2
        if self.object_type1 == src_type \
                and self.prop_name1 == ref_prop \
                and self.prop_name1 in src_obj \
                and self.object_type2 == dest_type \
                and self.prop_name2 in dest_obj:

            result = _refers_to(src_obj, self.prop_name1, dest_obj)

        # the reverse: check src/ref_prop matches with object/prop2
        elif self.object_type2 == src_type \
                and self.prop_name2 == ref_prop \
                and self.prop_name2 in src_obj \
                and self.object_type1 == dest_type\
                and self.prop_name1 in dest_obj:

            result = _refers_to(src_obj, self.prop_name2, dest_obj)

        else:
            result = False

        return result

    def is_applicable_types(self, src_type, ref_prop, dest_type):
        """
        Applicability check based only on types, not actual objects.  If an
        instance of src_type were linked to an instance of dest_type via the
        source object's ref_prop property, determine whether this constraint
        would constrain the dest instance (assuming it had the inverse
        property).

        :param src_type: An object type
        :param ref_prop: A property name
        :param dest_type: Another object type
        :return: True if an instance of dest_type would be constrained; False
            if not.
        """
        return (
            self.object_type1 == src_type
            and self.prop_name1 == ref_prop
            and self.object_type2 == dest_type
        ) or (
            self.object_type2 == src_type
            and self.prop_name2 == ref_prop
            and self.object_type1 == dest_type
        )

    def holds(self, src_obj, ref_prop, dest_obj):
        """
        Assuming applicability of this constraint (see is_applicable()), checks
        whether dest_obj refers back to src_obj via the inverse property (of
        ref_prop).

        :param src_obj: An object
        :param ref_prop: A property name
        :param dest_obj: Another object
        :return: True if the constraint holds; False if it does not
        """
        src_type = src_obj["type"]
        if self.object_type1 == src_type \
                and self.prop_name1 == ref_prop:
            result = _refers_to(dest_obj, self.prop_name2, src_obj)

        elif self.object_type2 == src_type \
                and self.prop_name2 == ref_prop:
            result = _refers_to(dest_obj, self.prop_name1, src_obj)

        else:
            result = False

        return result

    def enforce(self, src_obj, ref_prop, dest_obj):
        """
        Assuming applicability of this constraint (see is_applicable()),
        enforce this constraint on dest_obj.  This will force a property of
        dest_obj to refer to src_obj.  For a list-valued constraint property
        in dest_obj, an arbitrary element will be forced to src_obj's ID.

        :param src_obj: An object
        :param ref_prop: A property name
        :param dest_obj: Another object
        """

        if src_obj["type"] == self.object_type1 \
                and ref_prop == self.prop_name1:
            dest_prop_to_enforce = self.prop_name2
        else:
            dest_prop_to_enforce = self.prop_name1

        dest_prop_value = dest_obj[dest_prop_to_enforce]
        if isinstance(dest_prop_value, list):
            idx = random.randrange(len(dest_prop_value))
            dest_prop_value[idx] = src_obj["id"]

        else:
            dest_obj[dest_prop_to_enforce] = src_obj["id"]


_INVERSE_PROPERTIES = [
    InversePropertyConstraint(
        "network-traffic", "encapsulates_refs",
        "network-traffic", "encapsulated_by_ref"
    ),
    InversePropertyConstraint(
        "directory", "contains_refs", "file", "parent_directory_ref"
    ),
    InversePropertyConstraint(
        "process", "child_refs", "process", "parent_ref"
    )
]


class GraphType(enum.Enum):
    """
    Styles of graph:

    - TREE: No object "reuse", i.e. in-degree for all nodes is never greater
        than one.  Exceptions will occur for objects with fixed IDs, e.g. TLP
        marking definitions.
    - DAG: Directed acyclic graph: objects may be reused as long as no cycle is
        created.
    - RANDOM: Objects are randomly reused without regard to cycles.

    The acyclic graph topologies are still subject to inverse property cycles.
    So depending on how inverse property constraints are enforced, it may not
    be possible to have a truly acyclic graph.
    """
    TREE = 0
    DAG = 1
    RANDOM = 2


class InversePropertyConstraintEnforcement(enum.Enum):
    """
    An inverse property constraint is a constraint implied by two properties
    with an "inverse" relationship to each other.  (See the
    InversePropertyConstraint class.)  This enum determines how we should
    handle them.

    ENFORCE: Enforce them.  Where inverse properties occur, ensure they
        refer to each other's objects.  This means that acyclic graph
        topologies are not truly possible, because inverse properties will
        always result in a cycle.
    DELETE: Delete one of the properties.  This ensures inverse property
        cycles will not occur.  This is done naively, without regard to whether
        the property being deleted may have been required by spec.  So it could
        result in invalid objects, depending on how they were designed.
    IGNORE: Pretend there are no constraints.  This treats inverse properties
        as any other property.  This can result in nonsensical/contradictory
        semantics in the content.  E.g. directory1 contains file1, whose parent
        directory is directory2.  This means directory1 is declared to contain
        file1, but file1 claims to be in directory2.
    """
    ENFORCE = 0
    DELETE = 1
    IGNORE = 2


class Config(stix2generator.generation.Config):
    """
    Config settings for reference graph generation.

    - max_depth: about how many steps maximum from the "seed" object are we
        allowed to get.  This affects the maximum size of the graph.  It can't
        be strictly enforced, since one can't generally guarantee that a
        generated object will have no reference properties.
    - graph_type: what style of graph to create.  See the GraphType enum.
    - probability_reuse: when creating an object graph, in those situations
        where we can choose to reuse an already-created object, this determines
        how likely we are to choose to reuse.
    - inverse_property_constraints: how to handle inverse property constraints.
        See the InversePropertyConstraintEnforcement enum.
    """

    _DEFAULTS = {
        "max_depth": 3,
        "graph_type": GraphType.DAG,
        "probability_reuse": 0.5,
        "inverse_property_constraints":
            InversePropertyConstraintEnforcement.ENFORCE
    }


def _refers_to(src_obj, ref_prop, dest_obj):
    """
    Determine whether src_obj refers to dest_obj via reference property
    ref_prop.

    :param src_obj: A STIX object
    :param ref_prop: A reference property
    :param dest_obj: Another STIX object
    :return:
    """
    if ref_prop in src_obj:
        prop_value = src_obj[ref_prop]
        if isinstance(prop_value, list):
            result = dest_obj["id"] in prop_value
        else:
            result = prop_value == dest_obj["id"]

    else:
        result = False

    return result


def _is_reachable(src_id, dest_id, by_id, visited_ids=None):
    """
    Determine whether dest_id is reachable from src_id, following reference
    properties.

    :param src_id: A source object ID
    :param dest_id: A destination object ID
    :param by_id: The graph, as a mapping from ID to object
    :param visited_ids: A set which tracks visited IDs to avoid infinitely
        following cycles in the graph.  If None, a new set is automatically
        created and used.
    :return: True if dest_id is reachable from src_id; False if not
    """

    if src_id == dest_id:
        result = True

    elif src_id not in by_id:
        # Don't assume a "completed" ref graph, where every ref corresponds
        # to an existing graph node.  Be robust in case the graph is in a state
        # of partial construction and there are still some dangling references.
        result = False

    else:
        if visited_ids is None:
            visited_ids = set()

        if src_id in visited_ids:
            result = False
        else:
            visited_ids.add(src_id)

            src_obj = by_id[src_id]
            for _, ref_id in stix2generator.utils.find_references(src_obj):
                if _is_reachable(ref_id, dest_id, by_id, visited_ids):
                    result = True
                    break

            else:
                result = False

    return result


def _find_property_constraints(src_obj, ref_prop, dest_obj):
    """
    Generate InversePropertyConstraint instances applicable to src_obj and
    dest_obj, specifically with respect to the ref_prop property of src_obj.

    :param src_obj: An object
    :param ref_prop: A property name
    :param dest_obj: Another object
    """
    for constraint in _INVERSE_PROPERTIES:
        if constraint.is_applicable(src_obj, ref_prop, dest_obj):
            yield constraint


def _would_be_constrained(src_type, ref_prop, dest_type):
    """
    Determines whether if an object of type src_type hypothetically
    referred to an object of type dest_type via reference property
    ref_prop, would an inverse property constraint apply to the destination
    object and require it to be modified.  This is important for object reuse:
    if we want to reuse an object which has already had all its properties
    suitably set, we mustn't overwrite any of them to enforce an inverse
    property constraint.  Therefore, we can't reuse any object we'd have to
    apply a constraint to.

    I think this should be the case even if an instance of dest_type doesn't
    have the inverse property.  Graph semantics could still be wrong.  E.g.
    consider two directories who both say they contain the same file, where the
    file has no parent_directory_ref.  It doesn't matter whether the inverse
    property (parent_directory_ref) is present or not; it still doesn't make
    sense.

    :param src_type: A source object type
    :param ref_prop: A reference property name defined for src_obj_type
    :param dest_type: A destination object type
    :return: True if we'd have to apply a constraint to the destination object;
        False if not.
    """

    result = any(
        constraint.is_applicable_types(src_type, ref_prop, dest_type)
        for constraint in _INVERSE_PROPERTIES
    )

    return result


def _apply_constraints(src_obj, ref_prop, dest_obj):
    """
    Assuming src_obj refers to dest_obj via reference property ref_prop, apply
    any required inverse property constraints to dest_obj.

    :param src_obj: A source object
    :param ref_prop: A reference property name from src_obj
    :param dest_obj: A destination object to apply constraints to
    """
    constraints = _find_property_constraints(src_obj, ref_prop, dest_obj)
    for constraint in constraints:
        constraint.enforce(src_obj, ref_prop, dest_obj)


def _delete_inverse_properties(src_obj, ref_prop, dest_obj):
    """
    Assuming src_obj refers to dest_obj via reference property ref_prop,
    remove all inverse properties from dest_obj we would have to constrain, in
    order to avoid needing to apply any constraints.  This ensures there are no
    "back reference" properties pointing from dest_obj back to src_obj.

    :param src_obj: A source object
    :param ref_prop: A reference property name from src_obj
    :param dest_obj: A destination object to remove inverse properties from
    """
    constraints = _find_property_constraints(src_obj, ref_prop, dest_obj)

    for constraint in constraints:
        # the prop to delete is the inverse of ref_prop.
        if constraint.prop_name1 == ref_prop:
            prop_to_delete = constraint.prop_name2
        else:
            prop_to_delete = constraint.prop_name1

        dest_obj.pop(prop_to_delete, None)


class ReferenceGraphGenerator:
    """
    Instances of this class can generate object graphs such that the nodes
    are objects and edges correspond to reference properties of those
    objects.  These graphs do not use SROs.  Generation is designed to prevent
    any dangling references in the resulting graph.
    """

    def __init__(
        self, object_generator, config=None, stix_version="2.1"
    ):
        """
        Initialize this reference graph generator.

        :param object_generator: An instance of the built-in object generator
            to use for generating the objects.
        :param config: Config settings for the generator.
        """

        self.__config = config or Config()
        self.__object_generator = object_generator
        self.__stix_version = stix_version

        # A big problem that must be solved is how to stop the graph from
        # growing too large.  We can't simply stop generating objects, because
        # there must be no dangling references.  We also can't just keep
        # generating objects because the graph can balloon to enormous size.
        # My solution is to take advantage of the reference minimization
        # capability of the built-in object generator.  If we minimize
        # references, we minimize the number of new objects which need to be
        # generated, and that acts to put the brakes on growth.  We can't
        # minimize references all the time, because then the graph wouldn't
        # be able to grow much.  So we need two generators: one which generates
        # objects "normally" and is used to grow the graph, and one which
        # minimizes references, which is used to halt growth when we need it to
        # stop.

        halt_generator_config_dict = vars(object_generator.config).copy()
        halt_generator_config_dict["minimize_ref_properties"] = True
        halt_generator_config = stix2generator.generation.object_generator\
            .Config(
                **halt_generator_config_dict
            )
        self.__halt_generator = stix2generator.create_object_generator(
            halt_generator_config
        )

    @property
    def config(self):
        return self.__config

    def __random_generatable_seed_type(self, *constraints):
        """
        Pick a random seed type for a reference graph.

        :return: A random generatable seed STIX type
        :raises stix2generator.exceptions.GeneratableSTIXTypeNotFoundError: if
            a satisfying generatable STIX type could not be found
        """

        stix_type = stix2generator.utils.random_generatable_stix_type(
            self.__object_generator,
            *constraints,
            stix_version=self.__stix_version
        )

        if not stix_type:
            raise stix2generator.exceptions.GeneratableSTIXTypeNotFoundError(
                constraints, self.__stix_version
            )

        return stix_type

    def __augment_graph(self, object_, by_id, by_type, depth):
        """
        Build the graph out from the given object: find its reference
        properties and either create new objects for those properties to refer
        to, or choose other existing objects for them to refer to.

        :param object_: The graph object we are building out from
        :param by_id: The current graph, as a mapping from ID to object
        :param by_type: Another data structure containing the current graph
            nodes, as a mapping from STIX type name to a list of IDs of objects
            of that type.  This enables efficient lookup by type.
        :param depth: How many steps from the initial "seed" node the given
            object is.  Determines whether we need to stop graph growth.
        """

        if depth < self.config.max_depth:
            gen = self.__object_generator
        else:
            gen = self.__halt_generator

        for parent, key, ref_id, ref_prop \
                in stix2generator.utils.find_references_assignable(object_):

            # Constrained reference properties will refer to objects already
            # in the graph.  We must be careful not to disturb those.
            if ref_id in by_id:
                continue

            ref_type = stix2.utils.get_type_from_id(ref_id)
            new_object = None

            if self.config.graph_type is GraphType.TREE:
                # always generate a new object.
                new_object = gen.generate(ref_type)

            elif self.config.graph_type is GraphType.DAG:
                # Allow reuse if it doesn't create a cycle and it would not
                # require overwriting a property of the reused object due to
                # inverse property constraint enforcement.
                if (ref_type in by_type and
                        random.random() < self.config.probability_reuse and
                        (self.config.inverse_property_constraints is
                         InversePropertyConstraintEnforcement.IGNORE
                         or not _would_be_constrained(
                             object_["type"], ref_prop, ref_type
                         ))):
                    # Need to choose a random object which doesn't cause a
                    # cycle.  sample() here produces a new shuffled list,
                    # rather than shuffling in-place.  Shuffling means we can
                    # try each ID once, while at the same time not having a
                    # preference for IDs earlier in the original list.  The
                    # choice ought to be more random.
                    ids = by_type[ref_type]
                    shuffled_ids = random.sample(ids, k=len(ids))
                    for id_ in shuffled_ids:
                        if not _is_reachable(id_, object_["id"], by_id):
                            parent[key] = id_
                            break

                    else:
                        # All existing objects would create a cycle.  No
                        # choice but to create a new object.
                        new_object = gen.generate(ref_type)

                else:
                    new_object = gen.generate(ref_type)

            elif self.config.graph_type is GraphType.RANDOM:
                # Reuse randomly, not caring about cycles.
                if (ref_type in by_type and
                        random.random() < self.config.probability_reuse and
                        (self.config.inverse_property_constraints is
                         InversePropertyConstraintEnforcement.IGNORE
                         or not _would_be_constrained(
                            object_["type"], ref_prop, ref_type
                        ))):
                    id_ = random.choice(by_type[ref_type])
                    parent[key] = id_

                else:
                    new_object = gen.generate(ref_type)

            if new_object:

                # Replace the ref in the source obj with the dest object's ID,
                # rather than the other way around.  This is necessary since
                # some objects (e.g. certain marking definitions) are generated
                # with specific IDs they *must* use.  We mustn't replace them.
                new_object_id = new_object["id"]
                parent[key] = new_object_id

                if self.config.inverse_property_constraints is \
                        InversePropertyConstraintEnforcement.ENFORCE:
                    _apply_constraints(object_, ref_prop, new_object)
                elif self.config.inverse_property_constraints is \
                        InversePropertyConstraintEnforcement.DELETE:
                    _delete_inverse_properties(
                        object_, ref_prop, new_object
                    )
                # else IGNORE: leave new_object as-is

                by_id[new_object_id] = new_object
                by_type.setdefault(ref_type, []).append(new_object_id)

                self.__augment_graph(new_object, by_id, by_type, depth+1)

    def generate(self, seed_type=None, preexisting_objects=None):
        """
        Generate a reference graph seeded with an object of the given type.
        For graph types which allow reuse, preexisting_objects provides a way
        to reuse pre-existing objects.  This can be helpful when larger amounts
        of graph content is built up incrementally, so that new regions of
        content can connect with existing regions, instead of each invocation
        producing its own disconnected "island" of content.

        :param seed_type: A STIX type, STIXTypeClass enum, or None.  Generation
            is done by "building out" from a seed object.  This parameter
            determines the type of the seed object.  If None, a STIX SDO or SCO
            is chosen at random.
        :param preexisting_objects: The pre-existing STIX content, as either a
            list of objects or mapping from ID to object.
        :return: A 2-tuple including (1) The ID of the object which was created
            as the seed, and (2) the generated content, as a mapping from ID to
            object.  If preexisting_objects was given as a map, the same map is
            returned, updated with the new content.  Otherwise, a new map is
            created and returned.  It will contain all of the data from
            preexisting_objects plus the new content.
        """
        if not seed_type:
            seed_type = self.__random_generatable_seed_type(
                stix2generator.utils.STIXTypeClass.SDO,
                stix2generator.utils.STIXTypeClass.SCO
            )
        else:
            # Might seem kinda silly if seed_type is directly given as a string,
            # but this ensures that the given seed type is actually generatable
            # with our object generator.
            seed_type = self.__random_generatable_seed_type(seed_type)

        # Pre-populate our data structures, if we were given pre-existing
        # objects.
        if preexisting_objects is not None:
            if isinstance(preexisting_objects, collections.abc.Mapping):
                by_id = preexisting_objects
            else:
                by_id = {
                    obj["id"]: obj
                    for obj in preexisting_objects
                }
        else:
            by_id = {}

        # Wrapping the base preexisting objects map this way allows us to
        # easily distinguish new objects we create from old objects we were
        # given.  At the end, we only want to parse the new objects.
        by_id = collections.ChainMap(by_id).new_child()

        # For fast lookup of all generated objects of a particular type.  Maps
        # STIX type name to a list of object IDs of that type.
        by_type = {}

        for id_, obj in by_id.items():
            by_type.setdefault(obj["type"], []).append(id_)

        if self.config.max_depth == 0:
            gen = self.__halt_generator
        else:
            gen = self.__object_generator

        seed_object = gen.generate(seed_type)
        seed_id = seed_object["id"]

        by_type.setdefault(seed_type, []).append(seed_id)
        by_id[seed_id] = seed_object

        self.__augment_graph(seed_object, by_id, by_type, 0)

        # Parse the new objects we created and incorporate them into the base
        # graph.
        new_objects = by_id.maps[0]
        by_id = by_id.maps[1]
        for new_id, new_obj_dict in new_objects.items():
            by_id[new_id] = stix2.parse(
                new_obj_dict, version=self.__stix_version, allow_custom=True
            )

        return seed_id, by_id



import argparse
import configparser
import json

import stix2generator
import stix2generator.language.builder
import stix2generator.generation.object_generator
import stix2generator.logging


def parse_args():
    arg_parser = argparse.ArgumentParser(
        description="""Create STIX content from the STIX prototyping language"""
    )

    arg_parser.add_argument("language-file",
                            help="""
                            The file containing STIX prototyping language
                            """
                            )
    arg_parser.add_argument("-b", "--bundle",
                            help="Create a bundle",
                            action="store_true"
                            )
    arg_parser.add_argument("-e", "--encoding",
                            help="""
                            Encoding to use when reading text files, e.g.
                            STIX prototyping language, custom generator
                            specifications, etc.  Default=%(default)s
                            """,
                            default="utf-8"
                            )
    arg_parser.add_argument("-v", "--verbose",
                            help="""
                            Enable verbose diagnostic output.  Repeat for
                            increased verbosity.
                            """,
                            action="count"
                            )
    arg_parser.add_argument("--stix-version",
                            help="STIX version to use.  Default=%(default)s",
                            choices=["2.0", "2.1"],
                            default="2.1"
                            )
    arg_parser.add_argument("--extra-specs",
                            help="""A JSON file with extra object generator
                            specifications.  These will be merged with the
                            built-in specifications, and made available for use
                            in prototyping language content.
                            """
                            )
    arg_parser.add_argument("-n", "--embed-variable-names",
                            help="""Embed variable names in generated objects
                            using a custom property.
                            """,
                            action="store_true"
                            )
    arg_parser.add_argument("-c", "--config",
                            help="""Config file with options to customize how
                            content is generated.
                            """,
                            )

    args = arg_parser.parse_args()
    return args


def main():
    args = parse_args()

    stix2generator.logging.config_logging(args.verbose)

    proto_lang_file = getattr(args, "language-file")
    with open(proto_lang_file, "r", encoding=args.encoding) as f:
        proto_lang = f.read()

    extra_specs = None
    if args.extra_specs:
        with open(args.extra_specs, "r", encoding=args.encoding) as f:
            extra_specs = json.load(f)

    tmp_config = {}
    if args.config:
        config_parser = configparser.SafeConfigParser()
        config_parser.read(args.config)
        tmp_config = config_parser['main']

    generator_config = stix2generator.generation.object_generator.Config(
        **tmp_config
    )

    processor = stix2generator.create_default_language_processor(
        generator_config, extra_specs, args.stix_version
    )
    stix_objs = processor.build_graph(
        proto_lang, embed_variable_names=args.embed_variable_names
    )

    if args.bundle:
        bundle = stix2generator.utils.make_bundle(
            stix_objs, args.stix_version
        )

        print(bundle.serialize(pretty=True))

    else:
        for obj in stix_objs:
            print(obj.serialize(pretty=True))


if __name__ == "__main__":
    main()



import datetime
import hashlib
import logging
import random
import uuid

import pytz

import stix2generator.exceptions
import stix2generator.generation.pattern_generator
import stix2generator.generation.reference_graph_generator
import stix2generator.utils


# The property which names semantics.  Is "semantic" or "semantics" better?
SEMANTIC_PROPERTY_NAME = "semantics"


class SemanticsProvider:
    """
    Abstract base class for semantics providers.  Instances generate
    data given an abstract semantic name and some additional settings.

    NOTE: An implementation is provided the generator it was invoked through
    so that it may leverage auto-generation to create parts of its values.
    Avoid invoking a named specification from the given generator (e.g. via the
    "generate" method).  To be reusable across generators, a provider can't
    make assumptions as to which named specifications exist.  Also, you don't
    know what context the semantic is being invoked within: if the named
    specification is already in the midst of generation, you will cause
    infinite recursion.

    Semantics implementations should only generate from specifications which
    don't contain any references.
    """

    def __init__(self):

        # A logger for semantics providers, lazy-initialized
        self.__logger = None

    @property
    def logger(self):
        """
        Convenience property to get a logger named after the instance's module
        and class name.

        :return: A logger instance
        """
        if not self.__logger:
            cls = self.__class__
            self.__logger = logging.getLogger(
                cls.__module__ + "." + cls.__name__
            )

        return self.__logger

    def get_semantics(self):
        """
        Get a list of the semantics provided by the provider.  Subclasses
        must override this.

        :return: The list of semantics (list of strings which are the names
            of the semantics).
        """
        raise NotImplementedError()

    def create_semantic(self, spec, generator, constraint=None):
        """
        Create a value according to the given semantic spec.  For convenience,
        this default implementation tries to invoke a method named the same as
        the semantic name, except if the name contains hyphens.  If it contains
        hyphens, they are translated to underscores first, so it is easier to
        define the methods.  (Can't easily have a method name with a hyphen.)
        Subclasses can override this to do something else.

        :param spec: The semantic spec, which must be a mapping with a
            "semantics" property giving the name of the semantic.  Of course,
            that name should be one of the names returned by get_semantics()!
            Other properties may be used by the implementation, e.g. to
            influence how generation is done.
        :param generator: The generator through which this provider was
            invoked.  This allows implementations to delegate to other
            specifications as needed.  (But see the note in the class
            docstring.)
        :param constraint: A ValueConstraint instance representing some
            additional constraint to be honored by the generator.  This is
            derived from a co-constraint expression.  If None, there is no
            additional constraint.

        :return: The generated value
        """
        semantic = spec[SEMANTIC_PROPERTY_NAME].replace("-", "_")

        m = getattr(self, semantic)
        return m(spec, generator, constraint)


class FakerSemantics(SemanticsProvider):
    """
    A semantics provider which uses a faker.  The semantics names are taken
    directly from the faker object's providers' method names.  You can pass
    arguments into the faker method by adding additional properties to the
    specification.  They will be passed as keyword args.
    """

    def __init__(self, faker, faker_names=None):
        """
        Initialize this semantics provider.

        :param faker: The faker object to use
        :param faker_names: Iterable of names of faker methods to use from the
            given faker object, or None.  If None, the faker object's installed
            providers are scanned; this can result in a *lot* of names and a
            greater probability of semantic name collisions among semantics
            providers.
        """
        super().__init__()

        self.__faker = faker

        if faker_names:
            self.__semantics = list(faker_names)

            # Verify these names are valid faker methods?
            for name in self.__semantics:
                m = getattr(faker, name, None)
                if m is None:
                    # Or should this cause an exception...?
                    self.logger.warning(
                        "Faker function not found: %s", name
                    )

        else:

            # This is how faker finds methods from providers, as far as I could
            # tell from code inspection.  Faker also makes the methods
            # available directly on the faker object, but I still need to go
            # through the providers since a faker object has other methods
            # besides, and I wouldn't have a way of distinguishing.
            self.__semantics = [
                meth
                for prov in faker.get_providers()
                for meth in dir(prov)
                if not meth.startswith("_") and callable(getattr(prov, meth))
            ]

    def get_semantics(self):
        return self.__semantics

    def create_semantic(self, spec, generator, constraint=None):
        faker_kwargs = dict(spec)  # shallow-copy is ok

        # Remove stuff that's not part of the faker args, get desired
        # faker semantics name
        del faker_kwargs["type"]
        faker_semantics_name = faker_kwargs.pop(SEMANTIC_PROPERTY_NAME)

        faker_func = getattr(self.__faker, faker_semantics_name)
        return faker_func(**faker_kwargs)


class STIXSemantics(SemanticsProvider):
    """
    Some STIX-specific custom semantics.
    """

    SEMANTICS = [
        "stix-id",
        "stix-timestamp",
        "stix-pattern",
        "sha256",  # (conflicts with a Faker function)
        "sha512",
        "sha3_256",
        "sha3_512",
        "observable-container",
    ]

    _TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
    _TIMESTAMP_FORMAT_FRAC = "%Y-%m-%dT%H:%M:%S.%fZ"

    def get_semantics(self):
        return self.SEMANTICS

    def stix_id(self, spec, generator, constraint):
        """
        Create a STIX ID.  The spec requires the "stix-type" property, which
        determines the "<type>--" prefix.
        """
        if "stix-type" in spec:
            stix_type = spec["stix-type"]
            value = stix_type + "--" + str(uuid.uuid4())
        else:
            raise stix2generator.exceptions.ObjectGenerationError(
                'missing required "stix-type" property',
                spec[SEMANTIC_PROPERTY_NAME]
            )

        return value

    def stix_timestamp(self, spec, generator, constraint):
        """
        Create a random STIX timestamp which honors the given constraint.
        """
        if constraint:
            fmt = self._TIMESTAMP_FORMAT_FRAC if "." in constraint.value \
                else self._TIMESTAMP_FORMAT
            constraint_dt = datetime.datetime.strptime(constraint.value, fmt)
            constraint_dt = constraint_dt.replace(tzinfo=pytz.utc)

            if constraint.operator == constraint.EQ:
                timestamp_dt = constraint_dt

            else:
                equal_allowed = constraint.operator in (
                    constraint.GE, constraint.LE
                )

                # choose a random timestamp within a year of the constraint
                # timestamp...?
                random_duration = datetime.timedelta(
                    seconds=random.randint(
                        0 if equal_allowed else 1,
                        60 * 60 * 24 * 365 - 1
                    )
                )

                if constraint.operator in (constraint.LT, constraint.LE):
                    random_duration = -random_duration

                elif constraint.operator == constraint.NE:
                    if random.random() < 0.5:
                        random_duration = -random_duration

                # else for GE/GT, leave random_duration as-is

                timestamp_dt = constraint_dt + random_duration

        else:
            now_dt = datetime.datetime.now(tz=pytz.utc)

            # choose random within a year...?
            random_duration = datetime.timedelta(
                seconds=random.randint(
                    0,
                    60 * 60 * 24 * 365 - 1
                )
            )

            if random.random() < 0.5:
                random_duration = -random_duration

            timestamp_dt = now_dt + random_duration

        timestamp_str = timestamp_dt.strftime(self._TIMESTAMP_FORMAT)

        return timestamp_str

    def stix_pattern(self, spec, generator, constraint):
        """
        Create a random STIX pattern.
        """
        # Make kwargs from the spec, which we can use to create a pattern
        # generator Config object.  So configuring pattern generation is easy:
        # just use properties in the object generator spec which are named the
        # same as the Config settings for pattern generation.
        config_kwargs = dict(spec)
        config_kwargs.pop("type")
        config_kwargs.pop(SEMANTIC_PROPERTY_NAME)

        config = stix2generator.generation.pattern_generator.Config(
            **config_kwargs
        )
        patt_gen = stix2generator.generation.pattern_generator.PatternGenerator(
            generator, "2.1", config
        )

        pattern = patt_gen.generate()
        return pattern

    def _generate_hash(self, hash_function):
        """
        Generate a random hash using the given function.
        """
        value = hash_function(str(random.random()).encode())

        return value.hexdigest()

    def sha256(self, spec, generator, constraint):
        """
        Generate a random SHA256 hash.
        """
        return self._generate_hash(hashlib.sha256)

    def sha512(self, spec, generator, constraint):
        """
        Generate a random SHA512 hash.
        """
        return self._generate_hash(hashlib.sha512)

    def sha3_256(self, spec, generator, constraint):
        """
        Generate a random SHA3-256 hash.
        """
        return self._generate_hash(hashlib.sha3_256)

    def sha3_512(self, spec, generator, constraint):
        """
        Generate a random SHA3-512 hash.
        """
        return self._generate_hash(hashlib.sha3_512)

    def observable_container(self, spec, generator, constraint):
        """
        Generate an "observable container" object.  This is used as the value
        of the "objects" property of STIX 2.0-style observed-data objects.
        That property is deprecated in STIX 2.1.
        """

        config_kwargs = dict(spec)
        config_kwargs.pop("type")
        config_kwargs.pop(SEMANTIC_PROPERTY_NAME)

        config = stix2generator.generation.reference_graph_generator.Config(
            **config_kwargs
        )
        observable_container_generator = \
            stix2generator.generation.reference_graph_generator\
            .ReferenceGraphGenerator(generator, config)

        sco_type = stix2generator.utils.random_generatable_stix_type(
            generator, stix2generator.utils.STIXTypeClass.SCO
        )
        _, container = observable_container_generator.generate(sco_type)

        return container



#!/usr/bin/env python
from setuptools import setup, find_packages


def get_long_description():
    with open('README.rst') as f:
        return f.read()

setup(
    name='stix2-generator',
    version='0.1.0',
    description='Generate random STIX 2 content.',
    long_description=get_long_description(),
    long_description_content_type='text/x-rst',
    url='https://github.com/oasis-open/cti-stix-generator',
    author='OASIS Cyber Threat Intelligence Technical Committee',
    author_email='cti-users@lists.oasis-open.org',
    maintainer='Chris Lenk, Andy Chisholm',
    maintainer_email='clenk@mitre.org, chisholm@mitre.org',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='stix stix2 json generator generation stix2-generator',
    packages=find_packages(exclude=['*.test', '*.test.*']),

    python_requires='>= 3.5',

    install_requires=[
        'Faker',
        'lark-parser',
        'pytz',
        'stix2'
    ],

    extras_require={
        'jupyter': ['jupyter', 'stix2-viz'],
        'tests': ['pytest', 'rdflib']
    },

    package_data={
        'stix2generator': ['stix21_registry.json']
    },

    entry_points={
        'console_scripts': [
            'build_stix = stix2generator.language.build_stix:main',
            'generate_stix = stix2generator.generation.generate_stix:main'
        ]
    }
)
